# TODO try to make things more simple for ease of pickle problems

import importlib
import os
import sys
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from typing import Any
from typing import Union


def import_object(name: str) -> Any:
    """Import an object given a fully-qualified name.

    Args:
        - name (string): The fully-qualified name of the object to import.

    Returns:
        - obj: The object that was imported.

    Example:

    ```python
    >>> obj = import_object("random.randint")
    >>> import random
    >>> obj == random.randint
    True
    ```
    """
    try:
        mod_name, attr_name = name.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        return getattr(mod, attr_name)
    except ValueError:
        return importlib.import_module(name)


@contextmanager
def change_cwd(path: Union[str, Path]) -> Path:
    """
    A context manager which changes the working directory to the given
    path, and then changes it back to its previous _output on exit.
    Usage:
    > # Do something in original directory
    > with working_directory('/my/new/path'):
    >     # Do something in new directory
    > # Back to old directory
    """
    path = Path(path).expanduser().resolve()
    path.mkdir(parents=True, exist_ok=True)
    prev_cwd = os.getcwd()
    os.chdir(path)

    try:
        yield Path(path)
    finally:
        os.chdir(prev_cwd)


@contextmanager
def capture_local(event='return'):
    import sys
    local_vars = {}

    def capture_tracer(frame, _event, arg=None):
        if _event == event:
            local_vars.update(frame.f_locals)
        else:
            return capture_tracer

    sys.settrace(capture_tracer)

    yield local_vars

    sys.settrace(None)


def interrupt_main():
    import os
    if os.name == 'nt':
        from _thread import interrupt_main
        interrupt_main()
    else:
        import signal
        import threading
        signal.pthread_kill(threading.main_thread().ident, signal.SIGINT)


class CaptureTerminal(object):
    def __init__(self, stdout=None, stderr=None):
        if isinstance(stdout, str):
            stdout = open(stdout, 'a')
        if isinstance(stderr, str):
            stderr = open(stderr, 'a')
        self.stdout = stdout
        self.stderr = stderr
        self.stringio_stdout = StringIO()
        self.stringio_stderr = StringIO()

    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = self.stdout or self.stringio_stdout
        sys.stderr = self.stderr or self.stringio_stderr
        return self

    def __exit__(self, *args):
        self.stringio_stdout = StringIO()
        self.stringio_stderr = StringIO()
        sys.stdout = self._stdout
        sys.stderr = self._stderr



import asyncio
import uuid
from pathlib import Path
from typing import Union, List, Optional, Tuple

import flowsaber
from flowsaber.core.base import Component
from flowsaber.core.channel import Channel, Output
from flowsaber.core.task import BaseTask, Edge
from flowsaber.core.utils import check_cycle, class_deco
from flowsaber.server.database import FlowInput


class Flow(Component):
    """Represents the organizer of tasks, flows can also be used as components. Except for the top-most flow
    which represents the whole running unit, all flows within are simply virtual flows and don't have running state
    like flowrun or taskrun. However, flows and tasks all can have personalized configs.
    """
    FUNC_PAIRS = [('run', '__call__', True)]

    default_config = {
        'fork': 20,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.components: Optional[List[Component]] = None
        self.tasks: Optional[List[BaseTask]] = None
        self.edges: Optional[List[Edge]] = None

    @property
    def config_name(self) -> str:
        return "flow_config"

    def __enter__(self):
        flowsaber.context.flow_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        flowsaber.context.flow_stack.pop(-1)

    def call_initialize(self, *args, **kwargs):
        super().call_initialize(*args, **kwargs)
        # all flow record child executable components
        self.components = []
        # only the up-most flow record infos
        if not flowsaber.context.up_flow:
            self.tasks = []
            self.edges = []

    def initialize_context(self):
        """Expose some attributes of self.config into self.context.
        """
        # a task's workdir == context.flow_workdir / context.flow_config.workdir / context.task_config.workdir
        # If any of the above dirs is absolute dir, then use it instead
        super().initialize_context()
        # only inject the up-most flow's info
        if not flowsaber.context.up_flow:
            flow_workdir = self.config_dict['workdir']
            if not Path(flow_workdir).is_absolute():
                flow_workdir = str(Path().resolve().absolute())
            self.context.update({
                'flow_id': self.config_dict['id'],
                'flow_name': self.config_dict['name'],
                'flow_full_name': self.config_dict['full_name'],
                'flow_labels': self.config_dict['labels'],
                'flow_workdir': flow_workdir,
                'top_flow_config': self.config_dict
            })

    def call_build(self, *args, **kwargs) -> Union[Output, 'Flow']:
        # enter flow context
        from copy import deepcopy
        # used for child component
        self_context = deepcopy(self.context)
        self_context["up_" + self.config_name] = self_context[self.config_name]
        self_context.pop(self.config_name)
        with self, flowsaber.context(self_context):
            # TODO in case of None return, make a channel yield a end when all components run done
            self.output = self.run(*args, **kwargs) or Channel.end()

        # in order to check up_flow, can not run within with statement
        # now the dependency has been built
        if flowsaber.context.up_flow:
            flowsaber.context.up_flow.components.append(self)
            return self.output
        else:
            if check_cycle(self.task_id_edges):
                raise ValueError("The dependency graph has cycle.")
            # in the top level flow, just return the flow itself
            return self

    def run(self, *args, **kwargs) -> Channel:
        raise NotImplementedError("Not implemented. Users are supposed to compose flow/task in this method.")

    async def start_execute(self, **kwargs):
        """The up most flow needd to initialize executors.
        Parameters
        ----------
        kwargs

        Returns
        -------

        """
        await super().start_execute(**kwargs)

        async def execute_child_components():
            futures = []
            for component in self.components:
                future = asyncio.create_task(component.start(**kwargs))
                futures.append(future)

            # used fo debugging
            loop = asyncio.get_running_loop()
            if not hasattr(loop, 'flowsaber_futures'):
                loop.flowsaber_futures = []
            loop.flowsaber_futures += list(zip(self.components, futures))

            done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_EXCEPTION)
            # TODO wait for truely cancelled
            for fut in pending:
                fut.cancel()

            res_futures = list(done) + list(pending)
            self.check_future_exceptions(res_futures)

            return res_futures

        # for the top most flow, initialize executors
        if self.config_dict['id'] == self.context['flow_id']:
            flowsaber.context._info.clear()
            async with flowsaber.context:
                await execute_child_components()
                # for the top most flow, return None, since Flowrunner's returned final
                # state will inlcude this
        else:
            await execute_child_components()

    @property
    def task_id_edges(self) -> List[Tuple[str, str]]:
        edges = []
        for edge in self.edges:
            src_task = getattr(edge.channel, 'task', None)
            src_task_id = src_task.config_dict['id'] if src_task else str(uuid.uuid4())
            edges.append((src_task_id, edge.task.config_dict['id']))
        return edges

    def serialize(self) -> FlowInput:
        import base64
        import zlib
        from distributed.protocol.pickle import dumps
        # TODO can not fetch source code of type(self), if it's due to makefun ?
        assert self.initialized
        config = self.config
        compressed_flow = zlib.compress(dumps(self))
        serialized_flow = base64.encodebytes(compressed_flow).decode()
        return FlowInput(
            id=config.id,
            name=config.name,
            full_name=config.full_name,
            labels=config.labels,
            tasks=[task.serialize() for task in self.tasks],
            edges=[edge.serialize() for edge in self.edges],
            docstring=type(self).__doc__ or "",
            context=self.context,
            serialized_flow=serialized_flow
        )

    @classmethod
    def deserialize(cls, serialized_flow: str) -> "Flow":
        import base64
        import zlib
        from distributed.protocol.pickle import loads
        compressed_flow = base64.decodebytes(serialized_flow.encode())
        initialized_flow: 'Flow' = loads(zlib.decompress(compressed_flow))
        assert initialized_flow.initialized
        return initialized_flow


flow = class_deco(Flow, 'run')



import functools
import inspect
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Optional, Union, List, TYPE_CHECKING, Callable, Any

from makefun import with_signature
from pydantic import BaseModel

import flowsaber
from flowsaber.core.channel import Consumer, Output
from flowsaber.utility.context import Context, merge_dicts

if TYPE_CHECKING:
    import asyncio


def enter_context(method: Callable[..., Any]) -> Any:
    """A decorator runs the wrapped method within a new context composed of self.context and kwargs' context.

    Parameters
    ----------
    method

    Returns
    -------

    """

    @functools.wraps(method)
    def _enter_context(self, *args, **kwargs) -> Any:
        # we clear the package default context
        import flowsaber
        with flowsaber.context():
            with flowsaber.context(self.context):
                flowsaber.context.update(kwargs.get('context', {}))
                return method(self, *args, **kwargs)

    return _enter_context


def aenter_context(method: Callable[..., Any]) -> Any:
    """A decorator runs the wrapped method within a new context composed of self.context and kwargs' context.

    Parameters
    ----------
    method

    Returns
    -------

    """

    @functools.wraps(method)
    async def _aenter_context(self, *args, **kwargs) -> Any:
        import flowsaber
        with flowsaber.context():
            with flowsaber.context(self.context):
                flowsaber.context.update(kwargs.get('context', {}))
                return await method(self, *args, **kwargs)

    return _aenter_context


class ComponentExecuteError(RuntimeError):
    def __init__(self, *args, futures=None):
        super().__init__(*args)
        self.futures = futures or []


class ComponentCallError(RuntimeError):
    pass


class ComponentMeta(type):
    PAIR_ARG_NAME = 'FUNC_PAIRS'

    def __new__(mcs, class_name, bases, class_dict):

        class_name, bases, class_dict = mcs.copy_method_sig(class_name, bases, class_dict)
        class_name, bases, class_dict = mcs.update_default_config(class_name, bases, class_dict)

        cls = super().__new__(mcs, class_name, bases, class_dict)
        return cls

    @classmethod
    def copy_method_sig(mcs, class_name, bases, class_dict):
        """Used for automatically copy signature of a method to another method.
        Class must define 'FUNC_PAIRS' to indicate the src method and the target method

        FUNC_PAIRS:
            [src_method_name, target_method_name]
            [src_method_name, target_method_name, is_call_boolean]: signature of int will change to Channel[int]
        """

        # 1. handle copying method signature
        func_pairs = class_dict.get(mcs.PAIR_ARG_NAME, [])
        for base_cls in bases:
            base_func_pairs = getattr(base_cls, mcs.PAIR_ARG_NAME, [])
            func_pairs += base_func_pairs

        def copy_sig():
            src = class_dict.get(src_fn) or next(getattr(c, src_fn) for c in bases)
            tgt = class_dict.get(tgt_fn) or next(getattr(c, tgt_fn) for c in bases)
            while hasattr(src, '__inner_func__'):
                src = src.__inner_func__
            while hasattr(tgt, '__inner_func__'):
                tgt = tgt.__inner_func__
            src_sigs = inspect.signature(src)
            tgt_sigs = inspect.signature(tgt)
            # handle param signatures, this is used for run -> __call__
            if len(options) and options[0]:
                src_sig_params = list(src_sigs.parameters.values())
                for i, param in enumerate(src_sig_params):
                    if param.annotation is not inspect.Signature.empty:
                        src_sig_params[i] = param.replace(annotation=f"Channel[{param.annotation}]")
                src_sigs = inspect.Signature(src_sig_params, return_annotation=src_sigs.return_annotation)
            # handle return annotation, if tgt already has return annotation, keep it
            if tgt_sigs.return_annotation is not inspect.Signature.empty:
                src_sig_params = list(src_sigs.parameters.values())
                src_sigs = inspect.Signature(src_sig_params, return_annotation=tgt_sigs.return_annotation)

            @with_signature(src_sigs, func_name=tgt.__name__, qualname=tgt.__qualname__, doc=src.__doc__)
            def new_tgt_fn(*args, **kwargs):
                return tgt(*args, **kwargs)

            # used for source the real func
            new_tgt_fn.__source_func__ = src
            new_tgt_fn.__inner_func__ = tgt
            return new_tgt_fn

        if func_pairs:
            assert all(len(item) >= 2 for item in func_pairs)
            for src_fn, tgt_fn, *options in func_pairs:
                if src_fn == tgt_fn:
                    raise ValueError(f"src {src_fn} and tgt {tgt_fn} can not be the same.")
                new_tgt_fn = copy_sig()
                class_dict[tgt_fn] = new_tgt_fn

        return class_name, bases, class_dict

    @classmethod
    def update_default_config(mcs, class_name, bases, class_dict):
        """Automatically merge and update class-scoped `default_config` dict from parent class.
        Parameters
        ----------
        class_name
        bases
        class_dict

        Returns
        -------

        """
        # 2. handle default_config update
        from copy import deepcopy
        config_name = "default_config"
        default_config: dict = deepcopy(getattr(bases[0], config_name, {}))
        class_dict[config_name] = merge_dicts(default_config, class_dict.get(config_name, {}))

        return class_name, bases, class_dict


class Component(object, metaclass=ComponentMeta):
    """Base class of Flow and Task
    """

    class State(Enum):
        CREATED = 1
        INITIALIZED = 2
        EXECUTED = 3

    CREATED = State.CREATED
    INITIALIZED = State.INITIALIZED
    EXECUTED = State.EXECUTED

    default_config = {
        'id': None,
        'name': None,
        'full_name': None,
        'labels': [],
        'workdir': '',
        'log_stdout': True,
        'log_stderr': True,
    }

    def __init__(self, **kwargs):
        self.rest_kwargs = kwargs
        self.state: Component.State = self.CREATED
        self.context: Optional[dict] = None

        self._input_args: Optional[tuple] = None
        self._input_kwargs: Optional[dict] = None
        self._input: Optional[Consumer] = None
        self._output: Optional[Output] = None

    @property
    def input(self) -> Optional[Consumer]:
        return self._input

    @input.setter
    def input(self, value: Consumer):
        self._input = value

    @property
    def output(self) -> Optional[Output]:
        return self._output

    @output.setter
    def output(self, value: Output):
        self._output = value

    @property
    def config_name(self) -> str:
        raise NotImplementedError

    @property
    def config_dict(self) -> dict:
        if self.context is None:
            return {}
        else:
            return self.context[self.config_name]

    @property
    def config(self) -> Context:
        """return a non-editable context"""
        return Context(self.config_dict)

    @property
    def initialized(self):
        return self.state != Component.State.CREATED

    def __str__(self):
        name = None
        if self.initialized:
            name = self.config_dict.get('name')
        return name or f"{type(self).__name__}[{id(self)}]"

    def __repr__(self):
        full_name = None
        if self.initialized:
            full_name = self.config_dict.get("full_name")
        return full_name or str(self)

    def get_full_name(self) -> str:
        """Generate a name like flow1.name|flow2.name|flow3.name|cur_task
        """
        up_flow_names = '|'.join(flow.config_dict['name'] for flow in flowsaber.context.flow_stack)
        if up_flow_names:
            up_flow_names += '|'
        return f"{up_flow_names}{type(self).__name__}[{id(self)}]"

    def __call__(self, *args, **kwargs) -> Union[Output, 'Component']:
        """ This is where the flow/task build dependency graph

        Parameters
        ----------
        args
        kwargs

        Returns
        -------

        """
        try:
            from copy import copy
            new = copy(self)
            new.call_initialize(*args, **kwargs)
            build_output = new.call_build(*args, **kwargs)
            return build_output
        except BaseException as e:
            raise ComponentCallError(str(e)) from e

    def __copy__(self):
        cls = type(self)
        new = cls.__new__(cls)
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                new.__dict__[k] = None
            else:
                new.__dict__[k] = deepcopy(v)
        return new

    def call_initialize(self, *args, **kwargs):
        """Copy a new one and initialize some attributes.
        Parameters
        ----------
        args
        kwargs

        Returns
        -------

        """
        # copy a new one and initialize the context
        self.state = self.INITIALIZED
        self.initialize_context()

    def call_build(self, *args, **kwargs) -> Union[Output, 'Component']:
        raise NotImplementedError

    def initialize_context(self):
        """Called by call_initialize, merge and update self.config dict of self.context from different sources.
        """
        # may copied from a task already has context
        self.context = self.context or {}
        config_name = self.config_name
        up_workdir = flowsaber.context.get("up_" + config_name, {}).get('workdir', '')
        # update config in four steps
        default_config = self.default_config
        global_default_config = getattr(flowsaber.context, f'default_{config_name}', {})
        kwargs_config = self.rest_kwargs
        tmp_config = flowsaber.context.get(config_name, {})
        # initialize flow/task's context
        with flowsaber.context() as context:
            # 1: use class default config_dict
            context.update({config_name: default_config})
            # 2: use global default config_dict
            context.update({config_name: global_default_config})
            # 3. use kwargs settled config_dict
            context.update({config_name: kwargs_config})
            # 4. use user temporally settled config_dict
            context.update({config_name: tmp_config})
            context_dict = context.to_dict()
        self.context = merge_dicts(self.context, context_dict)

        # set up id, name, full_name
        if not self.config_dict.get('id'):
            self.config_dict['id'] = flowsaber.context.random_id
        if not self.config_dict.get('name'):
            self.config_dict['name'] = str(self)
        if not self.config_dict.get('full_name'):
            self.config_dict['full_name'] = self.get_full_name()

        # set up workdir build from the hierarchy of flows/tasks
        # flow's workdir and task's workdir is handled separately
        workdir = self.config_dict['workdir']
        workdir = Path(workdir).expanduser().resolve()
        if workdir.is_absolute():
            workdir = str(workdir)
        else:
            workdir = str(Path(up_workdir, workdir))
        self.config_dict['workdir'] = workdir

    def initialize_input(self, *args, **kwargs):
        self._input_args = args
        self._input_kwargs = kwargs

    # TODO should we use context manager
    @aenter_context
    async def start(self, **kwargs):
        """Start running the Flow/Task in the context of self.context, before setting the context,
        self.context will be merged/updated from kwargs.get('context', {})
        Parameters
        ----------
        kwargs

        Returns
        -------

        """
        # update context
        back_context = deepcopy(self.context)
        self.context = merge_dicts(self.context, kwargs.get('context', {}))
        try:
            res = await self.start_execute(**kwargs)
            return res
        except BaseException as e:
            raise ComponentExecuteError(str(e)) from e
        finally:
            await self.end_execute()
            self.context = back_context

    async def start_execute(self, **kwargs):
        if self.state == self.CREATED:
            raise ValueError("The Task/Flow object is not initialized, "
                             "please use task()/flow() to initialize it.")
        elif self.state == self.EXECUTED:
            raise ValueError("The Task/Flow has already been executed once before.")

        self.state = self.EXECUTED

    async def end_execute(self, *args, **kwargs):
        self.clean()

    def clean(self):
        pass

    def check_future_exceptions(self, futures: List['asyncio.Future']):
        # wrap into a ComponentExecuteError with record of all futures
        first_exception = next((fut.exception() for fut in futures if fut.exception()), None)
        if first_exception:
            raise ComponentExecuteError(str(first_exception), futures=futures) from first_exception

    def serialize(self) -> BaseModel:
        raise NotImplementedError

    def dict(self):
        return self.serialize().dict()



# this default context will only work in the build time
# during the flow/task run, will use context in flow/task as the initial context
DEFAULT_CONTEXT = {
    'default_flow_config': {
        'check_cancellation_interval': 5,
        'test__': {
            'test__': [1, 2, 3]
        },
        'log_stdout': True,
        'log_stderr': True,
    },
    'default_task_config': {
        'executor_type': 'dask',
        'test__': {
            'test__': [1, 2, 3]
        },
        'timeout': 0,
        'log_stdout': True,
        'log_stderr': True,
    },
    # logging options can be flow/task specific, here we treat it as global options for simplicity
    'logging': {
        'fmt': "[{levelname}] [{filename}:{lineno}-{funcName}()] "
               "[{task_full_name}] [{flowrun_id:1.6}|{taskrun_id:1.6}] : {message}",
        'datefmt': "%Y-%m-%d %H:%M:%S",
        'style': '{',
        'level': 'INFO',
        'buffer_size': 10,
        'context_attrs': [
            'flow_id',
            'flow_name',
            'flow_full_name',
            'task_id',
            'task_name',
            'task_full_name',
            'agent_id',
            'flowrun_id',
            'taskrun_id',
        ]
    },
    'caches': [
        {
            'cache_type': 'local'
        },
        {
            'cache_type': 'cloud',
            'address': 'xxxx',
            'cache_kwargs': None
        }
    ],
    'executors': [
        {
            'executor_type': 'local'
        },
        {
            'executor_type': 'dask',
            'address': None,
            'cluster_class': None,
            'cluster_kwargs': None,
            # TODO for best performance, use threads as worker, however, there may be bugs related to context and CWD
            # 'cluster_kwargs': {'threads_per_worker': 1},
            'adapt_kwargs': None,
            'client_kwargs': None,
            'debug': False
        }
    ]
}



from motor.motor_asyncio import AsyncIOMotorClient


def get_db(url: str, db_name: str = "flowsaber", **kwargs):
    client = AsyncIOMotorClient(url, **kwargs)
    db = client[db_name]

    return db


class DataBase(object):
    def __init__(self, url: str, db_name: str = "flowsaber", **kwargs):
        self.kwargs = kwargs
        self.url = url
        self.db_name = db_name
        self.db = None

    def __getattr__(self, item):
        if self.db is None:
            self.db = get_db(self.url, self.db_name, **self.kwargs)
        return getattr(self.db, item)

    @classmethod
    def test_connection(cls, url):
        import asyncio

        async def test():
            db = DataBase(url, serverSelectionTimeoutMS=1000)
            return await db.test_db.find_one()

        return asyncio.run(test())


if __name__ == "__main__":
    print(DataBase.test_connection('mongodb://127.0.0.1:27017'))



"""
Some functions come from prefect.utilities.context
"""
import contextlib
import contextvars
from collections import UserDict
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any, Iterable, Iterator, Union, cast

from flowsaber.core.default_context import DEFAULT_CONTEXT

DictLike = Union[dict, "DotBase"]


def merge_dicts(d1: DictLike, d2: DictLike) -> DictLike:
    """
    Updates `d1` from `d2` by replacing each `(k, v1)` pair in `d1` with the
    corresponding `(k, v2)` pair in `d2`.

    If the _output of each pair is itself a dict, then the _output is updated
    recursively.

    Args:
        - d1 (MutableMapping): A dictionary to be replaced
        - d2 (MutableMapping): A dictionary used for replacement

    Returns:
        - A `MutableMapping` with the two dictionary contents merged
    """

    new_dict = d1.copy()

    for k, v in d2.items():
        if isinstance(new_dict.get(k), MutableMapping) and isinstance(
                v, MutableMapping
        ):
            new_dict[k] = merge_dicts(new_dict[k], d2[k])
        else:
            new_dict[k] = d2[k]
    return new_dict


def as_nested_dict(
        obj: Union[DictLike, Iterable[DictLike]],
        dct_class: type = dict, **kwargs) -> Union[DictLike, Iterable[DictLike]]:
    """
    Given a obj formatted as a dictionary, transforms it (and any nested dictionaries)
    into the provided dct_class

    Args:
        - obj (Any): An object that is formatted as a `dict`
        - dct_class (type): the `dict` class to use (defaults to DotDict)

    Returns:
        - A `dict_class` representation of the object passed in
    ```
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)([as_nested_dict(d, dct_class) for d in obj])
    elif isinstance(obj, dct_class):
        for k, v in obj.items():
            obj[k] = as_nested_dict(v, dct_class)
    elif isinstance(obj, MutableMapping):
        # DotDicts could have keys that shadow `update` and `items`, so we
        # take care to avoid accessing those keys here
        return dct_class(
            {
                k: as_nested_dict(v, dct_class)
                for k, v in obj.items()
            },
            **kwargs
        )
    return obj


class DotBase(UserDict):
    PASS_ARGS = ['data']

    def __getattr__(self, item):
        return self[item]

    def __delattr__(self, item):
        del self[item]

    def __setattr__(self, key, value):
        if key.startswith('_') or key in self.PASS_ARGS:
            # __setattr__ prevail property.__set__
            if hasattr(type(self), key):
                object.__setattr__(self, key, value)
            else:
                self.__dict__[key] = value
        else:
            self[key] = value

    def to_dict(self) -> dict:
        """
        Converts current `DotDict` (and any `DotDict`s contained within)
        to an appropriate nested dictionary.
        """
        # mypy cast
        return cast(dict, as_nested_dict(self, dct_class=dict))


class DotDict(DotBase):
    """To make all levels of dict to be type(self), must use update to initialize
    """

    def __init__(self, *args, **kwargs: Any):
        # a DotDict could have a task_hash that shadows `update`
        super().__init__()
        # can not call super().__init__() as super().__init__() will again call self.update
        super().update(*args, **kwargs)

    def update(self, *args, **kwargs) -> None:
        dot_dict = dict(*args, **kwargs)
        # convert all dict into DotDict
        dot_dict = as_nested_dict(dot_dict, type(self))
        super().update(dot_dict)

    def __repr__(self):
        return f"{type(self).__name__}({self})"

    def __str__(self):
        return str(self.data)


class MergingDotDict(DotDict):
    def update(self, *args, **kwargs) -> None:
        # update dict recursively
        merged_dict = merge_dicts(self.data, dict(*args, **kwargs))
        super().update(merged_dict)


class Context(DotBase):
    """
    A coroutine safe context store


    Args:
        - *data (Any): arguments to provide to the `DotDict` constructor (e.g.,
            an initial dictionary)
        - **kwargs (Any): any task_hash / _output pairs to initialize this context with
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # The magic for ensuring coroutine safe
        self._data: contextvars.ContextVar = contextvars.ContextVar(f"data-{id(self)}")
        self._info: DotDict = DotDict()
        # the initial value of self._data is settled in here, within init will call self.data = dict(*args, **kwargs)
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"{type(self).__name__}({self.data})"

    def update(self, *args, **kwargs) -> None:
        self.data.update(*args, **kwargs)

    @property
    def data(self) -> MergingDotDict:
        try:
            return self._data.get()
        except LookupError:
            # this happens in thread, context var is not valid in new thread
            # dask uses thread to execute task
            # TODO Any other way?
            self.data = DEFAULT_CONTEXT
            return self._data.get()

    @data.setter
    def data(self, dic: DictLike):
        # this is the main point
        if not isinstance(dic, MergingDotDict):
            self._data.set(MergingDotDict())
            self.update(dic)
        else:
            self._data.set(dic)
        self._thread_backup = self._data.get()

    def __getstate__(self) -> None:
        """
        Because we dynamically update context during runs, we don't ever want to pickle
        or "freeze" the contents of context.  Consequently it should always be accessed
        as an attribute of the prefect module.
        """
        raise TypeError(
            "Pickling context objects is explicitly not supported. You should always "
            "access context as an attribute of the `flowsaber` module, as in `flowsaber.context`"
        )

    @contextlib.contextmanager
    def __call__(self, *args: MutableMapping, **kwargs: Any) -> Iterator["Context"]:
        """
        coroutine should use this context manager to update local context
        A context manager for setting / resetting the context context

        Example:
            from context import context
            with context(dict(a=1, b=2), c=3):
                print(context.a) # 1
        """
        # Avoid creating new `Context` object, copy as `dict` instead.
        update_dict = dict(*args, **kwargs)
        prev_context = self.to_dict()
        try:
            new_context = merge_dicts(deepcopy(prev_context), update_dict)
            self.data = new_context
            yield self
        finally:
            self.data = prev_context



import inspect
from typing import Any, Optional

import aiohttp
from pydantic import validate_arguments


class GraphqlError(Exception):
    pass


class ServerError(Exception):
    pass


class ValidateMeta(type):
    """Metaclass that automatically wrap method with pydantic.validate_arguemnts decorator.
    """

    def __new__(cls, clsname, bases, clsdict: dict):
        for k, v in clsdict.items():
            if not k.startswith('_') and inspect.isfunction(v):
                clsdict[k] = validate_arguments(v)
        return super().__new__(cls, clsname, bases, clsdict)


class Client(object, metaclass=ValidateMeta):
    """The Graphql client used for communicating with server by sending HTTP _post reqeuests data.
    """

    def __init__(self, server_url: str = None, retry: int = 3):
        self.server_url = server_url
        self.retry = retry
        self.test = True
        self.session: Optional[aiohttp.ClientSession] = None

    async def close(self):
        if self.session:
            await self.session.close()

    # json conflict with pydantic, so use underscore to pass validation
    async def _post(self, url, json=None, **kwargs) -> dict:
        if self.session is None:
            self.session = aiohttp.ClientSession()
        async with self.session.post(url, json=json, **kwargs) as rsp:
            return await rsp.json()

    async def graphql_request(self, query: str, variables: dict, **kwargs) -> dict:
        json = {
            'query': query,
            'variables': variables
        }
        # with multiple retry
        success = False
        retry = self.retry
        error = None
        rsp = {}
        while not success and retry > 0:
            try:
                rsp = await self._post(self.server_url, json=json, **kwargs)
                success = True
            except Exception as e:
                retry -= 1
                error = e
        # raise request error
        if not success and error:
            raise error
        # raise graphql error
        if 'errors' in rsp:
            raise GraphqlError(str(rsp['errors']))
        # raise server error
        if 'success' in rsp['data'] and not rsp['data']['success']:
            raise ServerError(str(rsp['data']))

        return rsp['data']

    async def query(self, method: str, input: Any, field: str):
        res = await self.graphql('query', method, input, field)
        return res

    async def mutation(self, method: str, input: Any, field: str):
        res = await self.graphql('mutation', method, input, field)
        return res

    async def graphql(self, method_type: str, method: str, input: Any, field: str):
        input_type = type(input).__name__
        if input_type == 'str':
            input_type = "UUID"

        query = f"""
            {method_type}($input: {input_type}!) {{
                {method}(input: $input) {{
                    {field}
                }}
            }}
        """
        variables = {
            'input': input.dict() if hasattr(input, 'dict') else input
        }

        result = await self.graphql_request(query, variables)
        return result[method]



import asyncio
from concurrent.futures import ProcessPoolExecutor, Future
from dataclasses import dataclass
from functools import partial
from typing import Set, Callable, Awaitable, Any, Sequence, Optional, Coroutine

AsyncFunc = Callable[[Any], Awaitable[None]]


class Solver(object):
    """Multiple knapsack problem solver.

    """

    def solve(self, *args, **kwargs):
        raise NotImplementedError


class GaSolver(Solver):
    """Use genetic algorithm supplied by pyeasyga package to solve the mkp problem.

    """

    def __init__(self, score_func: Callable = None):
        self.score_func = score_func

    def solve(self, items: Sequence) -> Sequence:
        from pyeasyga import pyeasyga
        if len(items) <= 0 or self.score_func is None:
            return items

        ga = pyeasyga.GeneticAlgorithm(list(items))
        ga.fitness_function = partial(self.fitness, self.score_func)
        ga.run()
        max_score, flags = ga.best_individual()

        return max_score, flags

    def fitness(self, score_func, individual, items: Sequence):
        assert len(items)
        total = None
        for selected, item in zip(individual, items):
            if selected:
                if total is None:
                    total = item
                else:
                    total += item
        return score_func(total)


class Scheduler(object):
    """Scheduler object should support create_task method that returns a Future object.

    """

    def create_task(self, *args, **kwargs) -> Future:
        raise NotImplementedError


class TaskScheduler(Scheduler):
    """A async task scheduler initialized with a scoring function. Each submited task is associated with a cost
    function, submited tasks will be selected and runned to ensure a maximum score.

    """

    @dataclass
    class Job(asyncio.Future):
        coro: Coroutine
        task: Optional[asyncio.Task] = None
        cost: Callable = None

        def cancel(self, *args, **kwargs) -> bool:
            if self.task:
                self.task.cancel(*args, **kwargs)
            return super().cancel(*args, **kwargs)

        def __post_init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __hash__(self):
            return hash(self.coro)

    def __init__(self, wait_time=1, score_func: Callable = None):
        self.pending_jobs: Set["TaskScheduler.Job"] = set()
        self.running_jobs: Set["TaskScheduler.Job"] = set()
        self.done_jobs: Set["TaskScheduler.Job"] = set()
        self.solver = GaSolver(score_func=score_func)
        self.error_jobs = []
        self.wait_time = wait_time
        self.loop_fut: Optional[Future] = None

        self._running: Optional[asyncio.Event] = None

    async def __aenter__(self):
        """Becarefull, there may be situation in which error raised in async with, but the loop has
        not been started yet. For this situation, we leave the async within only after the loop has been ran

        Returns
        -------

        """
        self._running = asyncio.Event()
        self.loop_fut = asyncio.ensure_future(self.start_loop())
        await self._running.wait()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._running.clear()
        await self.loop_fut
        self._running = None
        self.loop_fut = None

    def create_task(self, coro: Coroutine, cost: Callable = None) -> "TaskScheduler.Job":
        job = self.Job(coro, cost=cost)
        # add to tasks
        self.pending_jobs.add(job)
        # self.tasks[task].wait_q.put_nowait(1)
        return job

    async def start_loop(self, **kwargs):
        assert self._running is not None, "Please use `async with scheduler` statement to start the scheduler"
        self._running.set()

        while True:
            # use multiple knapsack problem sover to find the best subset of jobs given score_func
            try:
                # TODO sometimes this leads to error
                maydo = [job for job in self.pending_jobs if job.cost is not None]
                mustdo = [job for job in self.pending_jobs if job.cost is None]
                max_score, selected = self.solver.solve([job.cost() for job in maydo])
                jobs = [job for i, job in enumerate(maydo) if selected[i]] + mustdo
            except Exception:
                jobs = tuple(self.pending_jobs)

            for job in jobs:
                job = self.run_job(job)
            # break if not _running and no pending and _running jobs
            if not self._running.is_set() and len(self.running_jobs) + len(self.pending_jobs) == 0:
                break
            await asyncio.sleep(self.wait_time)

    def run_job(self, job: "TaskScheduler.Job"):
        async def _run_job():
            try:
                res = await job.coro
            except Exception as e:
                # record error, since the future are never been waited
                self.error_jobs.append(job)
                raise e
            finally:
                # always move to done
                self.running_jobs.remove(job)
                # self.done_jobs.add(job) # no need to store it
            return res

        # remove from pending to _running
        self.pending_jobs.remove(job)
        self.running_jobs.add(job)

        # propogate exception or result
        def run_job_done_callback(f: asyncio.Future):
            if f.cancelled():
                pass
            elif f.exception():
                job.set_exception(f.exception())
            elif f.done():
                job.set_result(f.result())
                # cancel should be triggered by job
            else:
                raise RuntimeError(f"Unexpected future {f}")

        job.task = asyncio.create_task(_run_job())
        job.task.add_done_callback(run_job_done_callback)

        return job


class FlowScheduler(Scheduler):
    """Internally it's a ProcessExecutor, the wrapping is mainly used for uniformity with TaskScheduler.
    """

    def __init__(self, executor_cls=ProcessPoolExecutor, executor_kwargs: dict = None):
        self.executor_cls = executor_cls
        self.executor_kwargs = executor_kwargs or {}
        self.executor: executor_cls = None
        self.futures: Set[Future] = set()

    def __enter__(self) -> 'FlowScheduler':
        self.executor = self.executor_cls(**self.executor_kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.executor.shutdown(wait=True)
        self.executor = None
        self.futures.clear()

    def create_task(self, *args, **kwargs) -> Future:
        fut = self.executor.submit(*args, **kwargs)
        fut.add_done_callback(self.remove_future)
        self.futures.add(fut)
        return fut

    def remove_future(self, fut):
        self.futures.remove(fut)



import sys
import threading
import time
from inspect import BoundArguments
from typing import TYPE_CHECKING

import flowsaber
from flowsaber.core.base import enter_context
from flowsaber.core.engine.runner import (
    Runner,
    catch_to_failure,
    call_state_change_handlers,
    run_timeout_signal,
    run_timeout_thread,
    redirect_std_to_logger
)
from flowsaber.core.utility.state import (
    State, Pending, Running, Retrying,
    Failure, Cached, Success, Skip, Drop
)
from flowsaber.core.utility.target import File
from flowsaber.server.database import TaskRunInput
from flowsaber.utility.statutils import ResourceMonitor

if TYPE_CHECKING:
    from flowsaber.core.task import Task


class TaskRunner(Runner):
    """The task runner moves the task state forward through a complicated process including:
    retry, cache read/write, skip, drop ...
    """

    def __init__(self, task: 'Task', inputs: BoundArguments, **kwargs):
        super().__init__(**kwargs)
        assert task.initialized
        self.task = task
        self.component = self.task
        self.inputs: BoundArguments = inputs

    def initialize_context(self, *args, **kwargs):
        update_context = {
            'taskrun_id': self.id,
            'server_address': self.server_address
        }
        self.context.update(**update_context)
        if 'context' in kwargs:
            kwargs['context'].update(**update_context)
        flowsaber.context.update(**update_context)

    @enter_context
    @redirect_std_to_logger
    @call_state_change_handlers
    @catch_to_failure
    def start_run(self, state: State = None, **kwargs) -> State:
        state = self.initialize_run(state, **kwargs)
        state = self.set_state(state, Pending)
        state = self.set_state(state, Running)
        # 1. skip if needed
        state = self.check_skip(state)
        if isinstance(state, Skip):
            return state
        retry = self.task.config_dict.get("retry", 1)
        cache_type = self.context.get('cache_type', None)
        while True:
            # 2. use cached result if needed
            if cache_type:
                state = self.read_cache(state)
                if isinstance(state, Cached):
                    return state
            # 3. run the task
            state = self.run_task(state, **kwargs)
            if isinstance(state, Failure):
                if retry > 0:
                    flowsaber.context.logger.warning(f"Run task: {self.task} failed, try to retry "
                                                     f"with {retry - 1} retrying left.")
                    state = self.set_state(state, Retrying)
                    time.sleep(self.task.config_dict.get('retry_wait_time', 2))
                    state = self.set_state(state, Running)
                    retry -= 1
                    continue
                elif self.task.config_dict.get("skip_error", False):
                    state = self.set_state(state, Drop)
            break
        # 4. write to cache if needed
        if isinstance(state, Success):
            if cache_type:
                state = self.write_cache(state)

        return state

    @call_state_change_handlers
    def check_skip(self, state: State) -> State:
        if self.task.skip_fn and self.task.need_skip(self.inputs):
            state = Skip.copy(state)
            state.result = tuple(state.inputs.arguments.values())
        return state

    @call_state_change_handlers
    def read_cache(self, state: State) -> State:
        run_workdir = self.context['run_workdir']
        no_cache = object()
        res = flowsaber.context.cache.get(run_workdir, no_cache)
        use_cache = res is not no_cache
        if use_cache:
            cache_valid = True
            values = [res] if not isinstance(res, (list, tuple, set)) else res
            for f in [v for v in values if isinstance(v, File)]:
                if f.initialized:
                    check_hash = f.calculate_hash()
                    if check_hash != f.hash:
                        msg = "Read cache failed from disk " \
                              "because file content task_hash changed."
                        flowsaber.context.logger.warning(msg)
                        cache_valid = False
                        break
            if cache_valid:
                state = Cached.copy(state)
                state.result = res
            else:
                flowsaber.context.cache.remove(run_workdir)
        return state

    @call_state_change_handlers
    @catch_to_failure
    def run_task(self, state: State, **kwargs) -> State:
        with ResourceMonitor() as monitor:
            res = self.run_task_timeout(**kwargs)
        state = Success.copy(state)
        state.result = res
        run_info = {
            'resource_usage': monitor.usage
        }
        self.context.update(run_info)
        flowsaber.context.update(run_info)

        return state

    @call_state_change_handlers
    def write_cache(self, state):
        assert isinstance(state, Success)
        run_workdir = self.context['run_workdir']
        cache = flowsaber.context.cache
        cache.put(run_workdir, state.result)
        cache.persist_single(run_workdir)
        return state

    def run_task_timeout(self, **kwargs):
        """Call task.run with timeout handling by using signal.
        Parameters
        ----------
        kwargs

        Returns
        -------

        """
        timeout = self.task.config_dict.get("timeout")

        if timeout:
            if not sys.platform.startswith('win'):
                is_main_thread = threading.current_thread() is threading.main_thread()
                if is_main_thread:
                    flowsaber.context.logger.debug("Run in timeout in main thread in unix system")
                    return run_timeout_signal(timeout, self.task.run, **self.inputs.arguments)

            flowsaber.context.logger.debug("Run in timeout using ThreadPoolExecutor")

            run_arguments = self.inputs.arguments

            @enter_context
            def run_in_context(task):
                return task.run(**run_arguments)

            return run_timeout_thread(timeout, run_in_context, self.task)

        return self.task.run(**self.inputs.arguments)

    def serialize(self, state: State, state_only=True):
        info = {'id': self.id, 'state': state.to_dict()}
        if not state_only:
            info.update({
                'flowrun_id': self.context['flowrun_id'],
                'agent_id': self.context.get('agent_id'),
                'task_id': self.context['task_id'],
                'flow_id': self.context['flow_id'],
                'context': {},
            })

        taskrun_input = TaskRunInput(**info)

        return taskrun_input



import asyncio

import flowsaber
from flowsaber.core.base import enter_context
from flowsaber.core.engine.runner import (
    Runner, catch_to_failure, call_state_change_handlers, redirect_std_to_logger
)
from flowsaber.core.engine.scheduler import TaskScheduler
from flowsaber.core.flow import Flow
from flowsaber.core.utility.state import State, Pending, Running, Success
from flowsaber.server.database.models import FlowRunInput


class FlowRunner(Runner):
    """Aimed for executing flow and maintaining/recording/responding state changes of the flow.
    """

    def __init__(self, flow: Flow, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(flow, Flow) and flow.initialized
        self.flow: Flow = flow
        self.component = self.flow
        if self.client:
            self.add_task(self.check_cancellation)
            self.add_task(self.maintain_heartbeat)

    def initialize_context(self, *args, **kwargs):
        update_context = {
            'flowrun_id': self.id,
            'server_address': self.server_address
        }
        self.context.update(**update_context)
        if 'context' in kwargs:
            kwargs['context'].update(**update_context)
        flowsaber.context.update(**update_context)

    @enter_context
    @redirect_std_to_logger
    @call_state_change_handlers
    @catch_to_failure
    def start_run(self, state: State = None, **kwargs) -> State:
        state = self.initialize_run(state, **kwargs)
        state = self.set_state(state, Pending)
        state = self.set_state(state, Running)
        state = self.run_flow(state, **kwargs)

        return state

    def leave_run(self, *args, **kwargs):
        super().leave_run()

    @call_state_change_handlers
    @catch_to_failure
    def run_flow(self, state, **kwargs):
        res = asyncio.run(self.async_run_flow(**kwargs))
        state = Success.copy(state)
        state.result = res
        return state

    async def async_run_flow(self, **kwargs):
        async with TaskScheduler() as scheduler:
            res = await self.flow.start(scheduler=scheduler, **kwargs)
        return res

    def serialize(self, state: State, state_only=True) -> FlowRunInput:
        info = {'id': self.id, 'state': state.to_dict()}
        if not state_only:
            info.update({
                'agent_id': self.context.get('agent_id'),
                'flow_id': self.context['flow_id'],
                'name': self.name,
                'labels': self.labels,
                'context': {},
            })

        flowrun_input = FlowRunInput(**info)

        return flowrun_input



import time
from threading import Thread, Event
from typing import Sequence

import psutil


class ResourceMonitor(Thread):
    """A resource monitor that starts another thread for monitoring resource usages within the `with` scope.
    """
    STATIC_ATTRS = ['num_ctx_switches', 'cpu_times']
    # STATIC_ATTRS = ['num_ctx_switches', 'io_counters', 'cpu_times']
    DYNAMIC_ATTRS = ['cpu_percent', 'num_threads', 'num_fds', 'memory_percent', 'memory_full_info']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.running = Event()
        self.d_usage = {}
        self.s_usage = {}
        self.num_count = 0
        self.usage = {}

    def __enter__(self):
        self.running.set()
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.running.clear()
        self.join()

    def run(self) -> None:
        self.s_usage = self.accum_resource(self.STATIC_ATTRS, self.s_usage)

        while self.running.is_set():
            time.sleep(1)
            self.accum_resource(self.DYNAMIC_ATTRS, self.d_usage)
            self.num_count += 1

        if self.num_count:
            for k, v in self.d_usage.items():
                self.d_usage[k] /= self.num_count

        s_usage = self.accum_resource(self.STATIC_ATTRS)
        for k, v in self.s_usage.items():
            s_usage[k] -= self.s_usage.get(k, 0)

        self.s_usage = s_usage

        self.usage = {**self.s_usage, **self.d_usage}

    @classmethod
    def accum_resource(cls, attrs: Sequence, usage=None):
        # TODO this method may cause errors sometimes, possibly due to quick exit of process.
        try:
            ps = [psutil.Process()]
            ps += list(ps[0].children(recursive=True))
            if usage is None:
                usage = {}

            for p in ps:
                p_usage = cls.unwrap(p.as_dict(attrs))
                for k, v in p_usage.items():
                    usage[k] = usage.get(k, 0) + v
            return usage
        except Exception:
            return {} if usage is None else usage

    @staticmethod
    def unwrap(value: dict):
        res = {}
        for k, v in value.items():
            if isinstance(v, (int, float)):
                res[k] = v
            elif hasattr(v, '__slots__') and len(v.__slots__):
                for sk in v.__slots__:
                    res[sk] = getattr(v, sk)
            elif hasattr(v, '__dict__') and len(v.__dict__):
                res.update(v.__dict__)
            elif hasattr(v, '_asdict'):
                res.update(v._asdict())
        return res



"""
Modified from prefect
"""
import logging
import sys
from logging import LogRecord
from logging.handlers import MemoryHandler
from typing import Tuple


class BufferHandler(MemoryHandler):

    def emit(self, record: LogRecord):
        msg = self.format(record)
        record.message = msg
        super().emit(record)


class LogHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handler = None

    def handle(self, record):
        handler = self.handler
        if handler is not None:
            handler(record)


def create_logger(name: str, log_record_factory, logging_options) \
        -> Tuple[logging.Logger, MemoryHandler, logging.Handler]:
    """
    comes from https://github.com/GangCaoLab/CoolBox/blob/master/coolbox/utilities/logtools.py
    """
    formatter = logging.Formatter(
        fmt=logging_options.fmt,
        datefmt=logging_options.datefmt,
        style=logging_options.style
    )

    logging.setLogRecordFactory(log_record_factory)
    logger = logging.getLogger(name)

    # stream logger use user defined log level
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging_options.level)
    stream_handler.setFormatter(formatter)

    # server log handler always capture all logs
    custom_log_handler = LogHandler()
    buffer_handler = BufferHandler(
        target=custom_log_handler,
        capacity=logging_options.buffer_size,
        flushLevel=logging.DEBUG
    )
    buffer_handler.setLevel(logging.DEBUG)
    buffer_handler.setFormatter(formatter)  # this has no effect,

    logger.addHandler(stream_handler)
    logger.addHandler(buffer_handler)
    logger.setLevel(logging.DEBUG)

    return logger, buffer_handler, custom_log_handler


class RedirectToLog:
    """
    Custom redirect of stdout messages to logs

    Args:
        - logger (logging.Logger, optional): an optional logger to redirect stdout. If
            not provided a logger names `stdout` will be created.
    """

    def __init__(self, logger: logging.Logger, level: int) -> None:
        self.logger: logging.Logger = logger
        self.level: int = level

    def write(self, msg: str) -> None:
        """
        Write message from stdout to a prefect logger.
        Note: blank newlines will not be logged.

        Args:
            msg (str): the message
        """
        if not isinstance(msg, str):
            # stdout is expecting str
            raise TypeError(f"string argument expected, got {type(msg)}")

        if msg.strip():
            self.logger.log(self.level, msg)

    def flush(self) -> None:
        """
        Implemented flush operation for logger handler
        """
        for handler in self.logger.handlers:
            handler.flush()



import shutil
from pathlib import Path
from typing import Any

from dask.base import tokenize

import flowsaber


class Serializer(object):
    def load(self, file):
        raise NotImplementedError

    def dump(self, value, file):
        raise NotImplementedError


class CloudPickleSerializer(Serializer):
    def load(self, file):
        from cloudpickle import load
        return load(file)

    def dump(self, value, file):
        from cloudpickle import dump
        return dump(value, file)


class CacheInvalidError(Exception):
    pass


NO_CACHE = object()


class Cache(object):
    """Cache is used for persisting results of met inputs. The cache should produce unique keys for unique inputs by
    implementing `hash` method. Hash of inputs can be further used to write/read related data.
    """

    def hash(self, **kwargs):
        raise NotImplementedError

    def put(self, input_key: str, output):
        raise NotImplementedError

    def get(self, input_key: str, default=NO_CACHE):
        raise NotImplementedError

    def remove(self, input_key: str):
        raise NotImplementedError

    def persist(self):
        raise NotImplementedError

    def persist_single(self, key):
        raise NotImplementedError


class LocalCache(Cache):
    """LocalCache treat hash keys as directories in disk and dump/load python objects in the corresponding directories.
    """
    VALUE_FILE = ".__Cache_value__"

    def __init__(self, serializer: Serializer = CloudPickleSerializer(), **kwargs):
        super().__init__(**kwargs)
        self.cache = {}
        self.serializer = serializer

    def __copy__(self):
        return self

    def __deepcopy__(self, memodict={}):
        return self

    def hash(self, **kwargs) -> str:
        h = tokenize(kwargs)
        return h

    def remove(self, key):
        path = Path(key)
        if path.exists():
            for f in path.glob('*'):
                if not f.name.startswith('._'):
                    if f.is_file():
                        f.unlink(missing_ok=True)
                    else:
                        shutil.rmtree(f, ignore_errors=True)

            return self.cache.pop(key, None)

    def get(self, key: str, default=NO_CACHE) -> object:
        # case 1, in memory cache
        if key in self.cache:
            flowsaber.context.logger.info(f"Read cache:{key} succeed from memory.")
        else:
            # case 2, in disk cache
            value_path = Path(key, self.VALUE_FILE)
            try:
                with value_path.open('rb') as f:
                    value = self.serializer.load(f)
            except Exception as e:
                self.remove(key)
                flowsaber.context.logger.debug(f"Read cache:{key} failed from disk with error: {e}")
                return default
            self.cache[key] = value
            flowsaber.context.logger.info(f"Read cache:{key} succeed from disk.")
        return self.cache[key]

    def put(self, key: str, data: Any):
        flowsaber.context.logger.debug(f"set cache:{key}")
        self.cache[key] = data

    def persist(self):
        """
        This should be called before python program ends
        """
        for key in tuple(self.cache.keys()):
            self.persist_single(key)

    def persist_single(self, key):
        flowsaber.context.logger.info(f"write cache {key}.")
        value = self.cache[key]
        f = Path(key, self.VALUE_FILE)
        f.parent.mkdir(parents=True, exist_ok=True)
        with f.open('wb') as wf:
            self.serializer.dump(value, wf)


def get_cache(cache: str = 'local', *args, **kwargs):
    cache_cls = {
        'local': LocalCache
    }
    if cache not in cache_cls:
        raise ValueError(f"cache type {cache} not supported, choose one of {cache_cls.keys()}")

    return cache_cls[cache](*args, **kwargs)



"""

During the running of flow, flowsaber.context will automatically be updated, Ideally, at the end of a task run,
context will be looked like this:

```
flow_config: {}
task_config: {}

flow_id
flow_name
flow_full_name
flow_labels

task_id
task_name
task_full_name
task_labels

id
flowrun_name

taskrun_id
```

"""

import asyncio
import inspect
import logging
import uuid
from collections import defaultdict
from contextlib import asynccontextmanager
from functools import partial
from typing import List, Optional, TYPE_CHECKING

from flowsaber.core.default_context import DEFAULT_CONTEXT
from flowsaber.core.utility.cache import Cache, get_cache
from flowsaber.core.utility.executor import Executor, get_executor
from flowsaber.utility.context import Context
from flowsaber.utility.logging import create_logger

if TYPE_CHECKING:
    from flowsaber.core.flow import Flow


class FlowSaberContext(Context):
    """The global coroutine-safe context meant to be used for inferring the running status of a flow at any time.
    Compared to raw context, this global context has intelligent(automatically change according to the
    time/position of the callee) properties:
        cache: used by running flow/task
        run_lock: used by running task
        logger: used in anywhere and anytime
        executor: used by running task

    """
    EXECUTOR_TABLE = '__executors'
    LOCK_TABLE = '__locks'
    CACHE_TABLE = "__caches"
    LOGGER_TABLE = '__loggers'
    FLOW_STACK_LIST = '__flow_stack'

    @property
    def random_id(self) -> str:
        return str(uuid.uuid4())

    @property
    def flow_stack(self) -> List['Flow']:
        return self._info.setdefault(self.FLOW_STACK_LIST, [])

    @property
    def top_flow(self) -> Optional['Flow']:

        return self.flow_stack[0] if self.flow_stack else None

    @property
    def up_flow(self) -> Optional['Flow']:
        return self.flow_stack[-1] if self.flow_stack else None

    # utility instance based on the current context
    async def __aenter__(self):
        """initialize executors
        Returns
        -------

        """
        # start executor
        executors = self._info.setdefault(self.EXECUTOR_TABLE, {})
        for executor_config in self.executors:
            executor_type = executor_config['executor_type']
            executors[executor_type] = get_executor(**executor_config)
        for executor in executors.values():
            self.logger.info(f"Starting executor: {executor}")
            await executor.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        executors = self._info[self.EXECUTOR_TABLE]  # must exists, or raise error
        # dask distributed has Warning about future was never awaited
        # https://github.com/dask/distributed/pull/3921
        for executor in executors.values():
            self.logger.info(f"Stopping executor: {executor}")
            await executor.__aexit__(exc_type, exc_val, exc_tb)
        executors.clear()
        del self._info[self.EXECUTOR_TABLE]

    @property
    def executor(self) -> Executor:
        """Fetch an initialized executor based on `executor_type` in the current context.
        Returns
        -------

        """
        assert 'executor_type' in self, "Can not find 'executor_type in context. You are not within a task."
        assert self.EXECUTOR_TABLE in self._info, 'Executors are not initialized, use async with statement ' \
                                                  'to initialize context'
        executors = self._info[self.EXECUTOR_TABLE]
        executor_type = self.executor_type
        if executor_type not in executors:
            self.logger.warning(f"The executor: {executor_type} not found. fall back to local")
        return executors[executor_type]

    @asynccontextmanager
    async def lock(self, keys: List[str]):
        """Fetch a asyncio.Lock based on `run_workdir` in the current context.
        Returns
        -------

        """
        lock_table = self._info.setdefault(self.LOCK_TABLE, defaultdict(asyncio.Lock))
        locks = tuple(lock_table[key] for key in keys)
        try:
            for lock in locks:
                await lock.acquire()
            yield locks
        finally:
            for lock in locks:
                lock.release()

    @property
    def cache(self, ) -> Cache:
        """Fetch a Cache based on `cache_type` in the current context.
        Returns
        -------

        """
        cache = self._info.setdefault(
            self.CACHE_TABLE,
            default=defaultdict(partial(get_cache, self.cache_type))
        )
        return cache[self.cache_type]

    @property
    def logger(self) -> logging.Logger:
        """Get a child logger of `flowsaber` logger with name of:
        `callee.__name__.agent_id.flow_id.id.task_id.taskrun_id`

        Returns
        -------

        """
        # find callee
        callee_frame = inspect.currentframe().f_back
        callee_module_name = callee_frame.f_globals['__name__']
        # TODO can add more running infos
        info_name = "NULL"

        logger_name = f"{callee_module_name}.{info_name}".rstrip('.')
        loggers = self._info.setdefault(self.LOGGER_TABLE, {})
        if logger_name not in loggers:
            loggers[logger_name] = flowsaber_logger.getChild(logger_name)

        return loggers[logger_name]


def inject_context_attrs(factory):
    """Inject context attrs into the log record base on context.context_attrs.
    Parameters
    ----------
    factory

    Returns
    -------

    """

    # TODO now users must explicitly specify attrs need to be attached to log_record
    # TODO Formatter._style.format has no ways to handle missing attts, thus causing errors
    def inner(*args, **kwargs):
        import flowsaber
        record = factory(*args, **kwargs)
        try:
            # TODO, sometimes, logging does not exist in flowsaber.context?
            context_attrs = flowsaber.context.logging.context_attrs or []
            for attr in context_attrs:
                setattr(record, attr, context.get(attr, ''))
        except Exception:
            pass
        return record

    return inner


context = FlowSaberContext()
context.update(DEFAULT_CONTEXT)

log_record_factory = logging.getLogRecordFactory()
log_record_factory = inject_context_attrs(log_record_factory)

# if use `logger` as name, will coflict with logger in dask
flowsaber_logger, buffer_handler, log_handler = create_logger("flowsaber", log_record_factory, context.logging)



import asyncio

import fire


class CLI:
    TEST_DB_URL = 'mongodb+srv://admin:admin@flowsaber.bkirk.mongodb.net/flowsaber?retryWrites=true&w=majority'

    def server(self, url: str = TEST_DB_URL, **kwargs):
        """

        Parameters
        ----------
        url: str
            The mongodb url. default: 127.0.0.1:27017
        kwargs: dict
            Options send to uvicorn.run()  for example: --host=xxx --port=xxx

        Returns
        -------

        """
        import uvicorn
        from flowsaber.server.app.app import get_app
        from flowsaber.server.database.db import DataBase
        # test connection
        DataBase.test_connection(url)
        # get awgi app
        db = DataBase(url)
        app = get_app(db)
        uvicorn.run(app=app, **kwargs)

    def agent(self, server: str, id: str = None, name: str = None, labels: list = None):
        """

        Parameters
        ----------
        server: str, optional
            The url of flowsaber server, You can run `flowsaber server MONGODB_URL` to get a server address.
        id: str
            The id of the current agent.
        name: str, optional
            The name of the current agent:
        labels: list, optional
            The labels attached to the current agent.

        Returns
        -------

        """
        import flowsaber
        from flowsaber.client.agent import Agent
        flowsaber.context.logger.info(f"Server: {server}")
        agent = Agent(server, id, name, labels)
        flowsaber.context.logger.info(f"The agent is: Agent(id={agent.id}, name={agent.name}, labels={agent.labels})")
        flowsaber.context.logger.info(f"Starting agent ....")
        asyncio.run(agent.start())


if __name__ == '__main__':
    fire.Fire(CLI)



from flowsaber.core.task import File
from flowsaber.tasks.shell import ShellFlow


class Bwa(ShellFlow):
    """
    For xxx.fastq.gz return a file named xx.fastq.bam
    """

    def command(self, fa: File, fastq: File, params=None):
        """
        bwa mem {params_str} -t {self.cpu} {fa} {fastq} | samtools view -Sbh -o {bam}
        """
        if params:
            assert isinstance(params, (tuple, list, str))
            if isinstance(params, str):
                params = (params,)
        params_str = ' '.join(params) if params else ''
        bam = f"{fastq.stem}.bam"
        return bam


bwa = Bwa(conda="bwa==0.7.12 samtools==1.2")



import asyncio
import builtins
from collections import defaultdict, abc
from typing import Callable, Any, Union

from flowsaber.core.channel import ConstantChannel, END, Consumer, LazyAsyncQueue, ChannelBase
from flowsaber.core.task import BaseTask
from flowsaber.core.utility.target import Data
from flowsaber.core.utils import extend_method, class_to_method

Predicate = Callable[[Any], bool]
POS_INF = float("inf")
NEG_INF = float("-inf")


class Operator(BaseTask):
    """Base class for all operators, subclass of BaseTask, all operators runs in the main loop in sequence
    and do not have runners and run states.
    """
    pass


class Merge(Operator):
    """Merge channels into a channel with _output of tuple.
    Even if there is only one channel _input, always _output a tuple
    """
    pass


class Split(Operator):
    """Used when _output is tuple/list, use split to split each item of the tuple into a unique channel.
    """

    def __init__(self, num: int, **kwargs):
        assert num >= 2, "Number of outputs must be at least 2."
        super().__init__(num_out=num, **kwargs)


class Select(Operator):
    """Similar to split, used for tuple/list _output, select will only _output one channel, with the selected index
    """

    def __init__(self, index: int, **kwargs):
        super().__init__(**kwargs)
        self.index = index

    async def handle_input(self, data, *args, **kwargs):
        if data is not END:
            try:
                await self.enqueue_res(data[self.index])
            except TypeError as e:
                raise ValueError(f"The output {data} can be be indexed with {self.index}. error is: {e}")


class Mix(Operator):
    """Data emitted bu channels are mixed into a single channel.
    """

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        num_constant_ch = builtins.sum(isinstance(q.ch, ConstantChannel) for q in consumer.queues)
        if num_constant_ch > 0:
            raise ValueError("Can not mix with constant channel.")

        async def pump_queue(q: LazyAsyncQueue):
            async for data in q:
                await self.enqueue_res(data)

        done, pending = await asyncio.wait(
            [asyncio.ensure_future(pump_queue(q)) for q in consumer.queues],
            return_when=asyncio.FIRST_EXCEPTION
        )
        for task in done:
            if task.exception() is not None:
                raise task.exception()


class Concat(Operator):
    """Data in channels are concatenated in the order of _input channels.
    """

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        for q in consumer.queues:
            async for data in q:
                await self.enqueue_res(data)


class Collect(Operator):
    """Opposite to flatten, turns a channel into a tuple
    """

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        values = []
        async for data in consumer:
            values.append(data)
        await self.enqueue_res(tuple(values))


class Branch(Operator):
    """Dispatch _input into specified number of channels base on the returned index of the predicate function.
    """

    def __init__(self, num: int, by: Callable, **kwargs):
        assert num > 1 and callable(by)
        super().__init__(num_out=num, **kwargs)
        self.fn = by

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        async for data in consumer:
            chosen_index = self.fn(data)
            if chosen_index >= self.num_out:
                raise RuntimeError(f"The returned index of {self.fn} >= number of _output channels")
            await self.enqueue_res(data, chosen_index)


class Sample(Operator):
    """Randomly sample at most `num` number of _input emitted by the channel
    using reservoir algorithm(should visit all elements.).
    """

    def __init__(self, num: int = 1, **kwargs):
        assert num >= 1, "Sample size must be at lest 1"
        super().__init__(**kwargs)
        self.num = num
        self.reservoir = []

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        import random
        cur = 0
        async for data in consumer:
            if cur < self.num:
                self.reservoir.append(data)
            else:
                i = random.randint(0, cur)
                if i < self.num:
                    self.reservoir[i] = data
            cur += 1
        for data in self.reservoir[0:cur]:
            await self.enqueue_res(data)


"""Map Task"""


class Subscribe(Operator):
    """specify on_next or on_complete function as callbacks of these two event.
    """

    def __init__(self, on_next: Callable = None, on_complete: Callable = None, **kwargs):
        super().__init__(**kwargs)
        self.on_next = on_next
        self.on_complete = on_complete

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        async for data in consumer:
            if self.on_next:
                self.on_next(data)
            await self.enqueue_res(data)
        if self.on_complete:
            self.on_complete()


class View(Subscribe):
    """Print each item emitted by the channel, Equals to call Subscribe(on_next=print)
    """

    def __init__(self, **kwargs):
        super().__init__(on_next=print, **kwargs)


class Map(Operator):
    """Map each item to another item return by the specified map function into a new channel.
    """

    def __init__(self, by: Callable, **kwargs):
        super().__init__(**kwargs)
        self.fn = by

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:
            res = self.fn(data)
            await self.enqueue_res(res)


class Reduce(Operator):
    """Similar to normal reduce. results = f(n, f(n - 1))
    """
    NOTSET = object()

    def __init__(self, by: Callable[[Any, Any], Any], result=NOTSET, **kwargs):
        super().__init__(**kwargs)
        self.fn = by
        self.result = result
        self.prev_result = self.NOTSET

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:
            result = self.fn(self.result, data)
            self.prev_result, self.result = self.result, result
        else:
            if self.prev_result is not self.NOTSET:
                await self.enqueue_res(self.result)
            self.result = self.NOTSET


class Flatten(Operator):
    """Flatten the _output of channel.
    """

    def __init__(self, max_level: int = None, **kwargs):
        super().__init__(**kwargs)
        self.max_level = max_level or 999999999999

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:

            flattened_items = []

            def traverse(items, level):
                try:
                    if level > self.max_level + 1:
                        raise TypeError
                    if isinstance(items, abc.Sequence) and not isinstance(items, str):
                        for _item in items:
                            traverse(_item, level + 1)
                    else:
                        raise TypeError
                except TypeError:
                    flattened_items.append(items)

            traverse(data, 1)

            for item in flattened_items:
                await self.enqueue_res(item)


class Group(Operator):
    """return a new channel with item of Tuple(key_fn(_input), grouped_data_tuple),
    the group size can be specified.
    """

    def __init__(self, by: Callable = lambda x: x[0], num: int = POS_INF, keep: bool = True, **kwargs):
        assert num > 1
        super().__init__(**kwargs)
        self.key_fn = by
        self.group_size = num
        self.keep_rest = keep
        self.groups = defaultdict(list)

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:
            key = self.key_fn(data)
            group = self.groups[key]
            group.append(data)
            if len(group) >= self.group_size:
                await self.enqueue_res((key, tuple(group)))
                del self.groups[key]
        else:
            if self.keep_rest:
                for key, values in self.groups.items():
                    await self.enqueue_res((key, tuple(values)))
                self.groups = defaultdict(list)


class Filter(Operator):
    """Filter item by the predicate function or the comparing identity.
    """

    def __init__(self, by: Union[Predicate, object] = lambda x: x, **kwargs):
        super().__init__(**kwargs)
        self.by = by

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:
            # two cases, filter by predicate or by _output
            keep = self.by(data) if callable(self.by) else data == self.by
            if keep:
                await self.enqueue_res(data)


class Unique(Filter):
    """Emit items at most once(no duplicate).
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cache = set()
        self.by = self.filter

    def filter(self, data):
        if data in self.cache:
            return False
        self.cache.add(data)
        return True


class Distinct(Filter):
    """Remove continuously duplicated item.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.by = self.filter
        self.prev = object()

    def filter(self, data):
        if data == self.prev:
            return False
        self.prev = data
        return True


class Take(Operator):
    """Only take the first `num` number of items
    """

    def __init__(self, num: int = 1, **kwargs):
        assert num >= 1
        super().__init__(**kwargs)
        self.num = num

    async def handle_input(self, data, *args, **kwargs):
        if data is not END and self.num > 0:
            await self.enqueue_res(data)
            self.num -= 1


class First(Take):
    """Take the first item"""

    def __init__(self, **kwargs):
        super().__init__(num=1, **kwargs)


class Last(Operator):
    """Take the last item"""
    NOTSET = object()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prev = self.NOTSET

    async def handle_input(self, data: Data, *args, **kwargs):
        if data is not END:
            self.prev = data
        else:
            if self.prev is not self.NOTSET:
                await self.enqueue_res(self.prev)
            self.prev = self.NOTSET


class Until(Operator):
    """"Take items until meet a stop marker. the stop item will not be included"""

    def __init__(self, by: Union[Predicate, object], **kwargs):
        super().__init__(**kwargs)
        self.by = by
        self.stop = False

    async def handle_input(self, data, *args, **kwargs):
        if data is not END and not self.stop:
            self.stop = self.by(data) if callable(self.by) else data == self.by
            if not self.stop:
                await self.enqueue_res(data)


class Min(Reduce):
    def __init__(self, result: float = POS_INF, **kwargs):
        super().__init__(by=min, result=result, **kwargs)


class Max(Reduce):
    def __init__(self, result: float = NEG_INF, **kwargs):
        super().__init__(by=max, result=result, **kwargs)


class Sum(Reduce):
    def __init__(self, **kwargs):
        super().__init__(by=sum, result=0, **kwargs)


class Count(Reduce):
    def __init__(self, **kwargs):
        super().__init__(by=lambda a, b: a + 1, result=0, **kwargs)


# though this more simple, but operators can not be detected by ide
# operators = set()
# operators_map = {}
for var in tuple(locals().values()):
    if isinstance(var, type) and issubclass(var, Operator) and var is not Operator:
        # operator = class_to_func(var)
        # operators_map[operator.__name__] = operator
        # operators.add(operator)
        extend_method(ChannelBase)(class_to_method(var))
# locals().update(operators_map)


merge = Merge()
flatten = Flatten()
mix = Mix()
concat = Concat()
view = View()
unique = Unique()
distinct = Distinct()
first = First()
collect = Collect()
last = Last()
count = Count()
min_ = Min()
sum_ = Sum()
max_ = Max()


def split(num: int):
    return Split(num=num)


def select(index: int):
    return Select(index=index)


def branch(num: int, by: Callable):
    return Branch(num=num, by=by)


def map_(by: Callable):
    return Map(by=by)


def reduce_(by: Callable[[Any, Any], Any], result=Reduce.NOTSET):
    return Reduce(by=by, result=result)


def filter_(by: Union[Predicate, object] = lambda x: x):
    return Filter(by=by)


def group(by: Callable = lambda x: x[0], num: int = POS_INF, keep: bool = True):
    return Group(by=by, num=num, keep=keep)


def subscribe(on_next: Callable = None, on_complete: Callable = None):
    return Subscribe(on_next=on_next, on_complete=on_complete)


def sample(num: int = 1):
    return Sample(num=num)


def take(num: int = 1):
    return Take(num=num)


def until(by: Union[Predicate, object]):
    return Until(by=by)



"""
Modified from https://github.com/PrefectHQ/prefect/blob/05cac2372c57a93ea72b05e7c844b1e115c01047/src
/prefect/tasks/notifications/email_task.py#L11
"""

import smtplib
import ssl
from email.message import EmailMessage
from typing import Any

from flowsaber.core.task import Task


class EmailTask(Task):
    """
    Task for sending email from an authenticated email service over SMTP. For this task to
    function properly, you must have the `"EMAIL_USERNAME"` and `"EMAIL_PASSWORD"` Prefect
    Secrets set.  It is recommended you use a [Google App
    Password](https://support.google.com/accounts/answer/185833) if you use Gmail.  The default
    SMTP server is set to the Gmail SMTP server on port 465 (SMTP-over-SSL). Sending messages
    containing HTML code is supported - the default MIME type is set to the text/html.
    Args:
        - subject (str, optional): the subject of the email; can also be provided at runtime
        - msg (str, optional): the contents of the email, added as html; can be used in
            combination of msg_plain; can also be provided at runtime
        - email_to (str, optional): the destination email address to send the message to; can also
            be provided at runtime
        - email_from (str, optional): the email address to send from; defaults to
            notifications@prefect.io
        - smtp_server (str, optional): the hostname of the SMTP server; defaults to smtp.gmail.com
        - smtp_port (int, optional): the port number of the SMTP server; defaults to 465
        - smtp_type (str, optional): either SSL or STARTTLS; defaults to SSL
        - msg_plain (str, optional): the contents of the email, added as plain text can be used in
            combination of msg; can also be provided at runtime
        - email_to_cc (str, optional): additional email address to send the message to as cc;
            can also be provided at runtime
        - email_to_bcc (str, optional): additional email address to send the message to as bcc;
            can also be provided at runtime
        - **kwargs (Any, optional): additional keyword arguments to pass to the base Task
            initialization
    """

    def __init__(
            self,
            username: str,
            password: str,
            subject: str = None,
            msg: str = None,
            email_to: str = None,
            email_from: str = "notifications@prefect.io",
            smtp_server: str = "smtp.gmail.com",
            smtp_port: int = 465,
            smtp_type: str = "SSL",
            msg_plain: str = None,
            email_to_cc: str = None,
            email_to_bcc: str = None,
            **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.username = username
        self.password = password
        self.subject = subject
        self.msg = msg
        self.email_to = email_to
        self.email_from = email_from
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.smtp_type = smtp_type
        self.msg_plain = msg_plain
        self.email_to_cc = email_to_cc
        self.email_to_bcc = email_to_bcc
        super().__init__(**kwargs)

    def run(self, *args, **kwargs) -> None:
        """
        Run method which sends an email.
        Args:
            Args will not be used.
        Returns:
            - None, Means the _output channel is a END channel
        """

        username = self.username
        password = self.password

        message = EmailMessage()
        message["Subject"] = self.subject
        message["From"] = self.email_from
        message["To"] = self.email_to
        if self.email_to_cc:
            message["Cc"] = self.email_to_cc
        if self.email_to_bcc:
            message["Bcc"] = self.email_to_bcc

        # https://docs.python.org/3/library/email.examples.html
        # If present, first set the plain content of the email. After which the html version is
        # added. This converts the message into a multipart/alternative container, with the
        # original text message as the first part and the new html message as the second part.
        if self.msg_plain:
            message.set_content(self.msg_plain, subtype="plain")
        if self.msg:
            message.add_alternative(self.msg, subtype="html")

        context = ssl.create_default_context()
        if self.smtp_type == "SSL":
            server = smtplib.SMTP_SSL(self.smtp_server, self.smtp_port, context=context)
        elif self.smtp_type == "STARTTLS":
            server = smtplib.SMTP(self.smtp_server, self.smtp_port)
            server.starttls(context=context)
        else:
            raise ValueError(f"{self.smtp_type} is an unsupported value for smtp_type.")

        server.login(username, password)
        try:
            server.send_message(message)
        finally:
            server.quit()



from pathlib import Path

from ariadne import (
    load_schema_from_path,
    make_executable_schema,
    asgi
)

from flowsaber.server.app.resolvers import get_resolvers
from flowsaber.server.database.db import DataBase

SCHEMA_PATH = Path(__file__).parent.resolve() / "graphql_schema"


def logging_post_data(resolver, obj, info, **kwargs):
    if obj is None:
        print(resolver, kwargs)
    return resolver(obj, info, **kwargs)


def get_app(db: DataBase):
    assert isinstance(db, DataBase)
    resolvers = get_resolvers(db)
    types = [resolvers[k] for k in ['query', 'mutation',
                                    'agent', 'flow', 'task', 'flowrun',
                                    'uuid_scalar', 'timestamp_scalar', 'json_scalar']]
    type_defs = load_schema_from_path(SCHEMA_PATH)
    schema = make_executable_schema(type_defs, *types)
    app = asgi.GraphQL(schema, middleware=[logging_post_data])
    return app



from collections import defaultdict
from typing import Union

from ariadne import QueryType, MutationType, ScalarType, ObjectType

from flowsaber.server.database.db import DataBase
from flowsaber.server.database.models import *


def ch_id(data: dict) -> dict:
    if "_id" in data:
        data['id'] = data.pop('_id')
    elif "id" in data:
        data['_id'] = data.pop("id")
    return data


def get_time_exp(input) -> dict:
    exp = {}
    before = getattr(input, 'before')
    after = getattr(input, 'after')
    if before or after:
        if after:
            exp['$gt'] = after
        if before:
            exp['$lt'] = before

    return exp


def update_notnone_exp(data: dict):
    # Note: does not check for list
    exp = {}

    def resolve(value, prevk=""):
        for k, v in value.items():
            if isinstance(v, dict):
                resolve(v, f"{k}.")
            elif v is not None:
                exp[f'{prevk}{k}'] = v

    resolve(data)
    exp.pop("id", None)
    exp.pop("_id", None)
    return {"$set": exp}


def get_field(obj: Union[BaseModel, dict], filed: str):
    if isinstance(obj, BaseModel):
        return getattr(obj, filed)
    else:
        return obj[filed]


def get_resolvers(db: DataBase):
    assert isinstance(db, DataBase)

    query = QueryType()
    mutation = MutationType()
    agent = ObjectType("Agent")
    flow = ObjectType("Flow")
    task = ObjectType("Task")
    flowrun = ObjectType("FlowRun")

    timestamp_scalar = ScalarType("TimeStamp")
    uuid_scalar = ScalarType("UUID")
    json_scalar = ScalarType("JSON")

    @timestamp_scalar.serializer
    def serialize_timestamp(value: float) -> float:
        assert isinstance(value, float)
        return value

    @uuid_scalar.serializer
    def serialize_uuid(value: str) -> str:
        assert isinstance(value, str)
        return value

    @json_scalar.serializer
    def serialize_json(value: dict) -> dict:
        assert isinstance(value, dict)
        return value

    # query

    @query.field('hello')
    async def hello(obj, info) -> str:
        return "Hello!"

    @query.field('get_agent')
    async def get_agent(obj, info, input: str) -> dict:
        agent_id = input
        agent_dict = await db.agent.find_one({"_id": agent_id})
        return ch_id(agent_dict)

    @query.field('get_agents')
    async def get_agents(obj, info) -> List[dict]:
        agents = []
        async for agent_dict in db.agent.find({}):
            agents.append(ch_id(agent_dict))

        return agents

    @query.field("get_flow")
    async def get_flow(obj, info, input: str) -> Flow:
        flow_id = input
        flow_dict = await db.flow.find_one({"_id": flow_id})
        flow_dict = ch_id(flow_dict)
        flow = Flow(**flow_dict)
        return flow

    @query.field("get_flows")
    async def get_flows(obj, info, input: dict) -> List[dict]:
        input = GetFlowsInput(**input)
        exp = {}
        if input.id or input.name or input.labels:
            exp.update({
                "$or":
                    [
                        {'_id': {'$in': input.id}},
                        {"name": {"$in": input.name}},
                        {"labels": {"$all": input.labels}}
                    ]

            })
        flows = []
        async for flow_dict in db.flow.find(exp):
            flows.append(ch_id(flow_dict))

        return flows

    @query.field("get_taskrun")
    async def get_taskrun(obj, info, input: str) -> TaskRun:
        taskrun_id = input
        taskrun_dict = await db.taskrun.find_one({"_id": taskrun_id})
        taskrun_dict = ch_id(taskrun_dict)
        taskrun = TaskRun(**taskrun_dict)
        return taskrun

    @query.field("get_taskruns")
    async def get_taskruns(obj, info, input: dict) -> List[dict]:
        input = GetTaskRunsInput(**input)
        exp = {}
        has_or_exp = input.id or input.task_id or input.flow_id or input.agent_id or input.flowrun_id
        if has_or_exp:
            exp.update({
                "$or":
                    [
                        {"_id": {"$in": input.id}},
                        {'task_id': {"$in": input.task_id}},
                        {"flow_id": {"$in": input.flow_id}},
                        {"agent_id": {"$in": input.agent_id}},
                        {"id": {"$in": input.flowrun_id}},
                    ]
            })
        if input.state_type:
            exp.update({
                "state.state_type": {"$in": input.state_type}
            })
        if input.before or input.after:
            time_exp = get_time_exp(input)
            exp.update({"start_time": time_exp})
        taskruns = []
        async for taskrun_dict in db.taskrun.find(exp):
            taskruns.append(ch_id(taskrun_dict))

        return taskruns

    @query.field("get_flowrun")
    async def get_flowrun(obj, info, input: str) -> FlowRun:
        flowrun_id = input
        flowrun_dict = await db.flowrun.find_one({"_id": flowrun_id})
        if flowrun_dict:
            flowrun_dict = ch_id(flowrun_dict)
            flowrun = FlowRun(**flowrun_dict)
        else:
            # for check_cancelling task, return a fake one
            flowrun = {'state': {'state_type': "Scheduled"}}
        return flowrun

    @query.field("get_flowruns")
    async def get_flowruns(obj, info, input: dict) -> List[dict]:
        input = GetFlowRunsInput(**input)
        exp = {}
        has_or_exp = input.id or input.flow_id or input.agent_id or input.name or input.labels
        if has_or_exp:
            exp.update({
                "$or":
                    [
                        {"_id": {"$in": input.id}},
                        {"flow_id": {"$in": input.flow_id}},
                        {"agent_id": {"$in": input.agent_id}},
                        {"name": {"$in": input.name}},
                        {"labels": {"$all": input.labels}},
                    ]
            })
        if input.state_type:
            exp.update({
                "state.state_type": {"$in": input.state_type}
            })
        if input.before or input.after:
            time_exp = get_time_exp(input)
            exp.update({"start_time": time_exp})

        flowruns = []
        async for flowrun_dict in db.flowrun.find(exp):
            flowruns.append(ch_id(flowrun_dict))

        return flowruns

    @query.field("get_runlogs")
    async def get_runlogs(obj, info, input: dict) -> List[dict]:
        input = GetRunLogsInput(**input)
        exp = {}
        has_or_exp = input.id or input.taskrun_id or input.flowrun_id or input.agent_id
        if has_or_exp:
            exp.update({
                "$or": [
                    {"_id": {"$in": input.id}},
                    {"taskrun_id": {"$in": input.taskrun_id}},
                    {"flowrun_id": {"$in": input.flowrun_id}},
                    {"agent_id": {"$in": input.agent_id}},
                ],
            })
        if input.level:
            exp.update({
                "level": {"$in": input.level}
            })
        if input.before or input.after:
            time_exp = get_time_exp(input)
            exp.update({"time": time_exp})

        runlogs = []
        async for runlog_dict in db.runlog.find(exp):
            runlogs.append(ch_id(runlog_dict))

        return runlogs

    # mutation

    @mutation.field("hello")
    async def resolve_write_hello(obj, info):
        return SuccessPayload()

    @mutation.field("create_agent")
    async def create_agent(obj, info, input: dict):
        agent_input = AgentInput(**input)
        request = info.context['request']
        address = request.client.host
        agent = Agent(**agent_input.dict(), address=address)
        await db.agent.delete_one({"_id": agent.id})
        await db.agent.insert_one(ch_id(agent.dict()))
        return agent

    @mutation.field("delete_agent")
    async def delete_agent(obj, info, input: str):
        agent_id = input
        res = await db.agent.delete_one({"_id": agent_id})
        return SuccessPayload(success=res.deleted_count == 1)

    @mutation.field("create_flow")
    async def create_flow(obj, info, input: dict):
        flow_input = FlowInput(**input)
        docs_dict = defaultdict(list)
        # store channels, tasks, flows
        for i, task_input in enumerate(flow_input.tasks):
            for j, ch_input in enumerate(task_input.output):
                task_input.output[j] = ch_input.id
                docs_dict['channel'].append(Channel(**ch_input.dict()))
            flow_input.tasks[i] = task_input.id
            docs_dict['task'].append(Task(**task_input.dict()))
        docs_dict['flow'].append(Flow(**flow_input.dict()))

        for collection, docs in docs_dict.items():
            docs = [ch_id(doc.dict()) for doc in docs]
            await getattr(db, collection).insert_many(docs)

        return docs_dict['flow'][0]

    @mutation.field("delete_flow")
    async def delete_flow(obj, info, input: str):
        flow_id = input
        res = await db.flow.delete_one({"_id": flow_id})
        return SuccessPayload(success=res.deleted_count == 1)

    @mutation.field("update_flowrun")
    async def update_flowrun(obj, info, input: dict):
        flowrun_input = FlowRunInput(**input)
        flowrun_id = flowrun_input.id
        flowrun = await db.flowrun.find_one({"_id": flowrun_id})
        if flowrun is None:
            # insert a new flowrun
            flowrun = FlowRun(**flowrun_input.dict())
            if not flowrun.start_time:
                flowrun.start_time = flowrun.last_heartbeat
            await db.flowrun.insert_one(ch_id(flowrun.dict()))
            # append to agent, flow 's flowruns
            await db.agent.update_one({"_id": flowrun.agent_id}, {"$push": {"flowruns": flowrun.id}})
            await db.flow.update_one({"_id": flowrun.flow_id}, {"$push": {"flowruns": flowrun.id}})
            return flowrun
        else:
            state_type = flowrun['state']['state_type']
            # in cancelling state, only allow cancelled new state
            if state_type == "Cancelling":
                if flowrun_input.state and flowrun_input.state.state_type != 'Cancelled':
                    return ch_id(flowrun)
            # in cancelled state, do not allow new state
            elif state_type == "Cancelled":
                return ch_id(flowrun)
            update_exp = update_notnone_exp(flowrun_input.dict())
            await db.flowrun.update_one({'_id': flowrun_id}, update_exp)
            updated_flowrun = await db.flowrun.find_one({"_id": flowrun_id})
            return ch_id(updated_flowrun)

    @mutation.field("update_taskrun")
    async def update_taskrun(obj, info, input: dict):
        taskrun_input = TaskRunInput(**input)
        taskrun_id = taskrun_input.id
        taskrun = await db.taskrun.find_one({"_id": taskrun_id})
        if taskrun is None:
            # insert a new task run
            taskrun = TaskRun(**taskrun_input.dict())
            if not taskrun.start_time:
                taskrun.start_time = taskrun.last_heartbeat
            await db.taskrun.insert_one(ch_id(taskrun.dict()))
            # append taskrun into the flowrun
            await db.flowrun.update_one({"_id": taskrun.flowrun_id}, {"$push": {"taskruns": taskrun.id}})
            return taskrun
        else:
            update_exp = update_notnone_exp(taskrun_input.dict())
            await db.taskrun.update_one({'_id': taskrun_id}, update_exp)
            updated_taskrun = await db.taskrun.find_one({"_id": taskrun_id})
            return ch_id(updated_taskrun)

    @mutation.field("write_runlogs")
    async def write_runlogs(obj, info, input: dict):
        runlogs_input = RunLogsInput(**input)
        run_logs = [ch_id(run_log.dict()) for run_log in runlogs_input.logs]
        res = await db.runlog.insert_many(run_logs)
        return SuccessPayload()

    #
    @flow.field("tasks")
    async def resolve_tasks(obj, info) -> List[dict]:
        task_ids = get_field(obj, 'tasks')
        tasks = await db.task.find({"_id": {"$in": task_ids}}).to_list(len(task_ids))
        tasks = [ch_id(task) for task in tasks]
        return tasks

    @task.field('output')
    async def resolve_channels(obj, info) -> List[dict]:
        channel_ids = get_field(obj, 'output')
        channels = await db.channel.find({"_id": {"$in": channel_ids}}).to_list(len(channel_ids))
        channels = [ch_id(channel) for channel in channels]
        return channels

    @flow.field("flowruns")
    @agent.field("flowruns")
    async def resolve_flowruns(obj, info) -> List[dict]:
        flowrun_ids = get_field(obj, 'flowruns')
        flowruns = await db.flowrun.find({"_id": {"$in": flowrun_ids}}).to_list(len(flowrun_ids))
        flowruns = [ch_id(flowrun) for flowrun in flowruns]
        return flowruns

    @flowrun.field("taskruns")
    async def resolve_taskruns(obj, info) -> List[dict]:
        taskrun_ids = get_field(obj, 'taskruns')
        taskruns = await db.taskrun.find({"_id": {"$in": taskrun_ids}}).to_list(len(taskrun_ids))
        taskruns = [ch_id(taskrun) for taskrun in taskruns]
        return taskruns

    return locals()



import builtins
import inspect
import types
from functools import partial
from typing import Union, Callable, List, Tuple, Any

from makefun import with_signature

from flowsaber.core.channel import ARGS_SIG, OUTPUT_ANNOTATION


def _self_func(self):
    pass


SELF_SIG = list(inspect.signature(_self_func).parameters.values())[0]


def get_sig_param(sig, param_type) -> tuple:
    return tuple(p for p in sig.parameters.values()
                 if p.kind == param_type)


def class_to_func(cls: type):
    """
    wrap:

        class A():
            def __init__(self, a: int, b: str = "x", **kwargs):
                pass
    into a function:
        def a(*data, a: int, b: str = "x", **kwargs):
            return A(a=a, b=b, **kwargs)(*data)
    """
    assert isinstance(cls, type), "The consumer argument must be a class"
    # Get signature of cls.__init__ except for self
    fn = types.MethodType(cls.__init__, object)
    fn_name = cls.__name__.lower()
    # check name collide with builtins
    if fn_name in dir(builtins):
        fn_name += "_by"
    # replace POSITIONAL_OR_KEYWORD to KEYWORD_ONLY
    # append *data VAR_POSITIONAL at the front
    sigs = list(inspect.signature(fn).parameters.values())
    for i, sig in enumerate(sigs):
        if sig.kind == inspect.Parameter.VAR_POSITIONAL:
            raise ValueError("The consumer cls.__init__ should not have *data: VAR_POSITIONAL parameter.")
        elif sig.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            sigs[i] = sig.replace(kind=inspect.Parameter.KEYWORD_ONLY)
    sigs.insert(0, ARGS_SIG)
    sigs = inspect.Signature(sigs, return_annotation=OUTPUT_ANNOTATION)

    @with_signature(sigs, func_name=fn_name, qualname=fn_name, doc=cls.__doc__)
    def inner(*args, **kwargs):
        return cls(**kwargs)(*args)

    return inner


def class_to_method(cls: type):
    """
    Wrap
        class A():
            def __init__(self, a: int, b: str = "x", **kwargs):
                pass
    into a function:
        def a(self, *data, a: int, b: str = "x", **kwargs):
            return A(a=a, b=b, **kwargs)(self, *data)
    """
    assert isinstance(cls, type), "The consumer argument must be a class"
    fn = types.MethodType(cls.__init__, object)
    fn_name = cls.__name__.lower()
    sigs = list(inspect.signature(fn).parameters.values())
    for i, sig in enumerate(sigs):
        if sig.kind == inspect.Parameter.VAR_POSITIONAL:
            raise ValueError(f"The input class {cls}.__init__ should not have *data: VAR_POSITIONAL parameter.")
        elif sig.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            sigs[i] = sig.replace(kind=inspect.Parameter.KEYWORD_ONLY)
    sigs = [SELF_SIG, ARGS_SIG] + sigs
    sigs = inspect.Signature(sigs, return_annotation=OUTPUT_ANNOTATION)

    @with_signature(sigs, func_name=fn_name, qualname=fn_name, doc=cls.__doc__)
    def inner(self, *args, **kwargs):
        return cls(**kwargs)(self, *args)

    return inner


def extend_method(cls):
    """Decorator to extend attributes of a class. Can be used in two ways:

    1:
    @extend_method(some_class)
    def new_method(self):
        pass

    2:
    @extend_method(some_class)
    class A:
        def new_method1(self):
            pass

        def new_method2(self):
            pass

    Parameters
    ----------
    cls

    Returns
    -------

    """

    def set_method(obj: Union[type, Callable]):
        import inspect

        funcs = []
        if inspect.isclass(obj):
            for name, func in inspect.getmembers(obj):
                if not name.startswith('_'):
                    funcs.append((name, func))
        elif inspect.isfunction(obj):
            funcs.append((None, obj))
        else:
            raise ValueError("Should be a class or function")

        for func_name, func in funcs:
            setattr(cls, func_name or func.__name__, func)

    return set_method


def class_deco(base_cls: type, method_name: str) -> Callable:
    def deco(fn: Callable = None, **kwargs) -> base_cls:
        """
        For base_cls is Task, method_name is run

        wrap  no-self argument function
            @deco
            def test(a, b, c) -> d:
                "doc"
                pass

        into a Class:
            class Test(Task):
                def run(self, a, b, c) -> d:
                    "doc"
                    return test(a, b, c)
        and return:
            Test()

        wrap with-self argument function
            @deco
            def test(self, a, b, c) -> d:
                "doc"
                pass

        into a Class:
            class Test(Task):
                def run(self, a, b, c) -> d:
                    "doc"
                    return test(self, a, b, c)
        and return:
            Test()
        """
        if fn is None:
            # TODO builtin partial does not maintain signature while makefun.partial has bug
            return partial(deco, **kwargs)
        cls_name: str = fn.__name__
        cls_name = cls_name[0].upper() + cls_name[1:]

        sig = inspect.signature(fn)
        sigs = list(sig.parameters.values())
        params = {
            'doc': fn.__doc__,
            'func_name': method_name,
            'qualname': method_name
        }
        no_args = not sigs
        first_is_self = sigs and sigs[0].name == 'self' and sigs[0].kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        if no_args or not first_is_self:
            sigs.insert(0, SELF_SIG)

            @with_signature(inspect.Signature(sigs, return_annotation=sig.return_annotation), **params)
            def wrapper(self, *args, **kwargs):
                return fn(*args, **kwargs)
        else:
            @with_signature(inspect.Signature(sigs, return_annotation=sig.return_annotation), **params)
            def wrapper(self, *args, **kwargs):
                return fn(self, *args, **kwargs)
        # used for source the real func
        wrapper.__source_func__ = fn
        cls = type(cls_name, (base_cls,), {method_name: wrapper})(**kwargs)
        return cls

    deco.__name__ = deco.__qualname__ = base_cls.__name__.lower()
    return deco


def check_cycle(edges: List[Tuple[Any, Any]]) -> bool:
    """Given a list of edges, check if the corresponding graph contains cycle by finding a topological sorting

    Parameters
    ----------
    edges

    Returns
    -------

    """
    from collections import defaultdict, deque
    # convert node to index
    nodes = set(node for edge in edges for node in edge)
    nodes2id = dict(zip(nodes, range(len(nodes))))
    edges = [(nodes2id[n1], nodes2id[n2]) for n1, n2 in edges]

    # build adjacency graph and in-degree vector
    G, indegree = defaultdict(list), [0] * len(nodes)
    for src, tgt in edges:
        indegree[tgt] += 1
        G[src].append(tgt)

    zero_dq = deque()
    for node, ind in enumerate(indegree):
        if ind == 0:
            zero_dq.append(node)
    # iteratively find zero in-degree node
    path = []
    while len(zero_dq):
        cur = zero_dq.popleft()
        path.append(cur)
        for out in G[cur]:
            indegree[out] -= 1
            if indegree[out] == 0:
                zero_dq.append(out)

    return len(path) != len(nodes)



import hashlib
from pathlib import Path
from typing import Union

import flowsaber


class Target(object):
    """Target represents item emitted by Channel. Theoretically all item emitted by Channel should be wrapped by a Target
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)


class File(Target):
    """Wrapping of `pathlib.Path`, with features including integrity checking ...
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj.path = Path(*args).expanduser().resolve()
        return obj

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self._hash_key = None

        if not self.path.is_file():
            flowsaber.context.logger.debug(f"File {self.path} does not exists.")

    def __getattr__(self, item):
        return getattr(self.path, item)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"File('{self}')"

    @property
    def hash(self) -> str:
        if not self._hash_key:
            raise ValueError("Should run File.calculate_hash before accessing hash_code.")
        return self._hash_key

    @hash.setter
    def hash(self, value: str):
        if self._hash_key:
            raise ValueError("Hash of file could only be settled one.")
        self._hash_key = value

    @property
    def initialized(self) -> bool:
        return self._hash_key is not None

    def calculate_hash(self) -> str:
        h = hashlib.md5()
        with self.open('rb') as f:
            # Read and update hash_code string _output in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                h.update(byte_block)
        return h.hexdigest()

    async def check_hash(self) -> bool:
        assert self.initialized
        new_hash = await flowsaber.context.executor.run(self.calculate_hash)
        return new_hash == self.hash

    def initialize_hash(self):
        if self.initialized:
            return
        new_hash = self.calculate_hash()
        self.hash = new_hash

    def __dask_tokenize__(self):
        return self.hash


class Stdout(File):
    """Ugly way, Use File to store stdout.
    In the idealist implementation, stdout and stdin can be piped/linked across machines over network.
    """


class Stdin(Target):
    """Represents stdin from a string or File.
    """

    def __init__(self, src: Union[str, File], **kwargs):
        assert isinstance(src, (str, File)), "Only string and File-like object can be converted to Stdin"
        super().__init__(**kwargs)
        self.src: Union[str, File] = src

    def __dask_tokenize__(self):
        if isinstance(self.src, str):
            return self.src
        else:
            return self.src.hash

    def __str__(self):
        if isinstance(self.src, str):
            return f"echo \"{self.src}\" | "
        else:
            return f"cat {self.src} | "


class End(Target):
    """End signal of a Channel.
    """

    def __repr__(self):
        return "[END]"

    def __copy__(self):
        return self

    def __hash__(self):
        return hash("[END]")


END = End()
Data = Union[tuple, End]



import asyncio
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING

import flowsaber
from flowsaber.client.client import Client
from flowsaber.core.engine.scheduler import FlowScheduler
from flowsaber.core.utility.state import Scheduled
from flowsaber.server.database.models import AgentInput, GetFlowRunsInput

if TYPE_CHECKING:
    from concurrent.futures import Future


# default ProcessPoolExecutor can not pickle build flow
def run_flowrun(flowrun_data: dict, flow_data: dict, server_address: str, agent_id: str):
    from flowsaber.api import Flow, FlowRunner, State

    flow_run_kwargs = {
        'context': {
            **flowrun_data['context'],
            'agent_id': agent_id
        }
    }

    try:
        state = State.from_dict(flowrun_data['state'])
        flow = Flow.deserialize(flow_data['serialized_flow'])
        flow_runner = FlowRunner(
            flow,
            server_address=server_address,
            id=flowrun_data['id'],
            name=flowrun_data['name'],
            labels=flowrun_data['labels']
        )
        flow_runner.run(state=state, **flow_run_kwargs)

    except BaseException as e:
        raise RuntimeError(f"Run failed, with error: {e}.")


class Agent(object):
    """Agent fetches available flowruns in Scheduled state, and use FlowScheduler and Flowrunner
    for executing flow in processes.
    """

    def __init__(self, server: str, id: str = None, name: str = None, labels: list = None):
        self.id = id or flowsaber.context.random_id
        self.name = name or flowsaber.context.random_id
        self.labels = labels or []
        self.client = Client(server)
        self.server = server

        self.flowruns = defaultdict(dict)

    async def start(self):
        flowsaber.context.logger.info(f"Started aagent: {self.__dict__}")
        agent_input = AgentInput(
            id=self.id,
            name=self.name,
            labels=self.labels
        )
        await self.client.mutation('delete_agent', self.id, "success")
        agent_data = await self.client.mutation('create_agent', agent_input, "id")
        assert agent_data['id'] == self.id
        with FlowScheduler() as scheduler:
            # retrieve scheduled flowruns endlessly
            while True:
                # fetch flowrun ids
                get_flowruns_input = GetFlowRunsInput(
                    state_type=[Scheduled().state_type],
                    agent_id=[self.id]
                )
                flowruns = await self.client.query(
                    'get_flowruns',
                    get_flowruns_input,
                    "id name labels flow_id state {state_type result message} context"
                )
                for flowrun in flowruns:
                    flowrun_id = flowrun['id']
                    if flowrun_id in self.flowruns:
                        flowsaber.context.logger.debug(f"Find already executed flowrun: {flowrun_id}, Pass.")
                        continue
                    else:
                        flowsaber.context.logger.info(f"Fetched scheduled flowrun: {flowrun_id} "
                                                      f"for flow: {flowrun['flow_id']}, try for running...")
                    try:
                        # fetch flow
                        flow = await self.client.query('get_flow', flowrun['flow_id'],
                                                       "serialized_flow")
                        # use flow_runner to run the flow

                        fut = scheduler.create_task(
                            run_flowrun,
                            flowrun,
                            flow,
                            self.server,
                            self.id
                        )
                        flowrun_info = self.flowruns[flowrun_id]
                        flowrun_info.update({
                            'flow': flow,
                            'future': fut,
                        })

                        fut.add_done_callback(partial(self.report, flowrun_id))

                    except Exception as e:
                        flowsaber.context.logger.info(f"Meet error when fetching/executing "
                                                      f"flowrun:{flowrun_id} 's flow with error: {e}")
                # sleep for some time
                await asyncio.sleep(3)

    def report(self, flowrun_id: int, fut: 'Future'):
        logger = flowsaber.context.logger
        if fut.cancelled():
            logger.info(f"The flowrun: {flowrun_id} has been cancelled, remove it from flowruns.")
            self.flowruns.pop(flowrun_id)
        elif fut.exception():
            logger.info(f"The flowrun: {flowrun_id} run failed with exception: {fut.exception()}")
        else:
            logger.info(f"The flowrun: {flowrun_id} run succeed with return: {fut.result()}")


if __name__ == "__main__":
    agent = Agent("http://127.0.0.1:8123", id="test")

    asyncio.run(agent.start())



import asyncio
import inspect
from collections import abc
from inspect import Parameter, BoundArguments
from pathlib import Path
from typing import Callable, Sequence, Optional, TYPE_CHECKING, List

from dask.base import tokenize

import flowsaber
from flowsaber.core.base import Component, aenter_context
from flowsaber.core.channel import Channel, Consumer, Output
from flowsaber.core.engine.task_runner import TaskRunner
from flowsaber.core.utility.state import State, Done, Success, Failure
from flowsaber.core.utility.target import File, END, Data
from flowsaber.core.utils import class_deco
from flowsaber.server.database.models import EdgeInput, TaskInput

if TYPE_CHECKING:
    from flowsaber.core.engine.scheduler import TaskScheduler


class RunDataTypeError(TypeError):
    pass


class RunDataFileNotFoundError(RuntimeError):
    pass


class BaseTask(Component):
    """Base class of all Tasks, internally, BaseTask iteratively fetch items emitted by Channel inputs asynchronously.
    And then push the processed result of each item into the _output channels. All items are handled in sequence.
    """
    default_config = {
        'workdir': 'work'
    }

    def __init__(self, num_out: int = 1, **kwargs):
        super().__init__(**kwargs)
        self.num_out = num_out

    @property
    def config_name(self) -> str:
        return "task_config"

    def initialize_context(self):
        """Initialize some attributes of self.config into self.context
        """
        super().initialize_context()
        # only expose the info of the outermost flow
        self.context.update({
            'task_id': self.config_dict['id'],
            'task_name': self.config_dict['name'],
            'task_full_name': self.config_dict['full_name'],
            'task_labels': self.config_dict['labels'],
        })

    def call_build(self, *args, **kwargs) -> Output:
        with flowsaber.context(self.context):
            self.initialize_input(*args, **kwargs)
            self.initialize_output()
        # enter task context
        # register to top flow for serializing
        top_flow = flowsaber.context.top_flow
        top_flow.tasks.append(self)
        # register to up flow for execution
        up_flow = flowsaber.context.up_flow
        up_flow.components.append(self)

        return self.output

    def initialize_input(self, *args, **kwargs):
        """Wrap all _input channels into a consumer object for simultaneous data ferching,

        Parameters
        ----------
        args
        kwargs
        """
        super().initialize_input(*args, **kwargs)
        channels = list(args) + list(kwargs.values())
        self.input = Consumer.from_channels(*channels, task=self)
        # register edges into the up-most flow
        for input_q in self.input.queues:
            edge = Edge(channel=input_q.ch, task=self)
            try:
                flowsaber.context.top_flow.edges.append(edge)
            except Exception as e:
                raise e

    def initialize_output(self):
        """Create _output channels according to self.num_output
        """
        self.output = tuple(Channel(task=self) for i in range(self.num_out))
        if self.num_out == 1:
            self.output = self.output[0]

    async def start_execute(self, **kwargs):
        await super().start_execute(**kwargs)
        await self.handle_consumer(self.input, **kwargs)
        # always sends a END to _output channel
        end_signal = END if self.num_out == 1 else [END] * self.num_out
        await self.enqueue_res(end_signal)

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        """Iteratively fetch data from consumer and then call processing function

        Parameters
        ----------
        consumer
        kwargs
        """
        async for data in consumer:
            await self.handle_input(data)
        await self.handle_input(END)

    async def handle_input(self, data, *args, **kwargs):
        """Do nothing, send the _input data directly to _output channel
        Parameters
        ----------
        data
        args
        kwargs
        """
        if data is not END:
            await self.enqueue_res(data)

    async def enqueue_res(self, data, index=None):
        """Enqueue processed data into _output channels.
        Parameters
        ----------
        data
        index
        """
        # enqueue data into the _output channel
        if self.num_out != 1 and isinstance(self.output, Sequence):
            try:
                if index is None:
                    for ch, _res in zip(self.output, data):
                        await ch.put(_res)
                else:
                    await self.output[index].put(data)
            except TypeError as e:
                raise RuntimeError(f"The _output: {data} can't be split into {self.num_out} channels."
                                   f"The error is {e}")
        else:
            await self.output.put(data)

    def __ror__(self, chs) -> Output:
        """
        ch | task               -> task(ch)
        [ch1, ch2, ch3] | task  -> task(ch1, ch2, ch3)
        """
        if not isinstance(chs, abc.Sequence):
            chs = [chs]
        assert all(isinstance(ch, Channel) for ch in chs)
        return self(*chs)

    def __rrshift__(self, chs):
        """
        ch >> task              -> task(ch)
        [ch1, ch2, ch3] >> task -> [task(ch1), task(ch2), task(ch3)]
        """
        if isinstance(chs, abc.Sequence):
            assert all(isinstance(ch, Channel) for ch in chs)
            output_chs = [self(ch) for ch in chs]
            if isinstance(chs, tuple):
                output_chs = tuple(output_chs)
            return output_chs
        else:
            assert isinstance(chs, Channel)
            return self(chs)

    def __lshift__(self, chs):
        """
        task << ch
        task << [ch1, ch2, ch3]
        """
        return chs >> self

    @classmethod
    def source_code(cls) -> str:
        import inspect
        return inspect.getsource(cls)

    @classmethod
    def input_signature(cls) -> dict:
        call_signature = {
            param: str(param_type)
            for param, param_type
            in inspect.signature(cls.__call__).parameters.items()
        }

        return call_signature

    def serialize(self) -> TaskInput:
        # TODO can not fetch source code of type(self), if it's due to makefun ?
        config = self.config
        output = self.output
        if self.num_out == 1:
            output = [self.output]
        return TaskInput(
            id=config.id,
            flow_id=self.context['flow_id'],
            name=config.name,
            full_name=config.full_name,
            labels=config.labels,
            output=[ch.serialize() for ch in output],
            docstring=type(self).__doc__ or "",
            context=self.context
        )


class RunTask(BaseTask):
    """RunTask is subclass of BaseTask, representing tasks with run method exposed to users to implement specific
    item processing logics.
    Compared to BaseTask:
    1. Runs of multiple inputs will be executed in parallel.
    2. Runs will be executed in the main loop.
    """
    FUNC_PAIRS = [('run', '__call__', True)]
    default_config = {
        'drop_error': False,
        'retry': 0,
        'fork': 7,
        'cpu': 1,
        'gpu': 0,
        'memory': 0.2,
        'time': 1,
        'io': 1,
        'cache_type': 'local',
        'executor_type': 'dask',
        'timeout': 0
    }

    def initialize_context(self):
        """Expose cache_type and executor_type into self.context
        """
        super().initialize_context()
        self.context.update({
            'cache_type': self.config_dict['cache_type'],
            'executor_type': self.config_dict['executor_type']
        })

    async def handle_consumer(self, consumer: Consumer, **kwargs):
        """Run processing functions in parallel by submitting jobs into schedulers that
        return awaitable Future-like objects.
        Parameters
        ----------
        consumer
        kwargs

        Returns
        -------

        """
        # get the custom scheduler and pop it, do not pass into task runner
        scheduler: 'TaskScheduler' = kwargs.get("scheduler", None)
        kwargs.pop('scheduler', None)

        futures = []
        async for data in consumer:
            run_data = (data,) if self.input.single else data
            # split run data and dependent data
            run_data = self.create_run_data(run_data)
            job_coro = self.handle_run_data(run_data, **kwargs)
            # add_task to scheduler
            if scheduler:
                fut = scheduler.create_task(job_coro)
            else:
                fut = asyncio.create_task(job_coro)
            futures.append(fut)
        # asyncio.wait can not accept empty list
        if not futures:
            return []

        # wait for the first exception and cancel all functions
        done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_EXCEPTION)
        for fut in pending:
            # TODO wait for truely cancelled
            fut.cancel()

        res_futures = list(done) + list(pending)
        self.check_future_exceptions(res_futures)

        return res_futures

    def create_run_data(self, data: Data) -> BoundArguments:
        """Wrap consumer fetched data tuple into a BoundArgument paired with self.run's signature.
        Parameters
        ----------
        data

        Returns
        -------

        """
        # 1. build BoundArgument
        len_args = len(self._input_args)
        args = data[:len_args]
        kwargs = {
            k: data[len_args + i] for i, k in enumerate(self._input_kwargs.keys())
        }

        run_data = inspect.signature(self.run).bind(*args, **kwargs)
        run_data.apply_defaults()
        return run_data

    @aenter_context
    async def handle_run_data(self, data: BoundArguments, **kwargs):
        """This coroutine will be executed in parallel, thus need to re-enter self.context.
        Parameters
        ----------
        data
        kwargs
        """
        data: BoundArguments = await self.check_run_data(data, **kwargs)
        res = await self.call_run(data, **kwargs)
        await self.handle_res(res)

    async def check_run_data(self, data: BoundArguments, **kwargs):
        """Match types of _input datas into self.run's annotations by type conversion. Check file integrity.
        Parameters
        ----------
        data
        kwargs

        Returns
        -------

        """
        run_params = dict(inspect.signature(self.run).parameters)
        arguments = data.arguments
        for arg, param in run_params.items():
            ano_type = param.annotation
            if ano_type is not Parameter.empty and isinstance(ano_type, type):
                value = arguments[arg]
                # 1. do some type conversion in case with type annotation
                is_default = param.default is not inspect.Signature.empty and value == param.default
                if not is_default and not isinstance(value, ano_type):
                    try:
                        arguments[arg] = ano_type(value)
                    except Exception as e:
                        raise RunDataTypeError(f"The input argument `{arg}` has annotation `{ano_type}`, "
                                               f"but the input value `{value}` can not be converted.") from e
                # 2. if has File annotation, make sure it exists
                if ano_type is File and not arguments[arg].is_file():
                    raise RunDataFileNotFoundError(f"The argument {arg} has a File annotation, "
                                                   f"but the file {value} does not exists.")
            # 3. make sure each File's  checksum being computed, only check the first level
            value = arguments[arg]
            values = [value] if not isinstance(value, (list, tuple, set)) else value
            for f in [v for v in values if isinstance(v, File)]:
                if not f.initialized:
                    new_hash = await flowsaber.context.executor.run(f.calculate_hash)
                    f.hash = new_hash
        return data

    async def call_run(self, data: BoundArguments, **kwargs):
        """Create a fresh task object and call it's run method for real data processing.
        """
        from copy import copy
        clean_task = copy(self)
        res = clean_task.run(**data.arguments)
        return res

    async def handle_res(self, res):
        await self.enqueue_res(res)

    def run(self, *args, **kwargs):
        """The method users need to implement for processing the data emited by _input channels.
        Parameters
        ----------
        args
        kwargs
        """
        raise NotImplementedError("Please implement this method.")


class Task(RunTask):
    """Task is subclass of RunTask:
    1. Each Task will have a unique task_key/task_workdir
    2. Each _input's run will have a unique run_key/task_workdir.
    3. Task's run will be executed in executor and handled by a task runner.
    4. Within the task runner, task will pass through a state machine, callbacks can be registered to each state changes.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.skip_fn: Optional[Callable] = None

    @property
    def task_hash(self) -> str:
        """get the task_hash of this task defined by the real source code of self.run

        Returns
        -------

        """
        fn = self.run
        # for wrapped func, use __source_func__ as a link
        while hasattr(fn, '__source_func__'):
            fn = fn.__source_func__
        code = fn.__code__.co_code
        annotations = getattr(fn, '__annotations__', None)

        return tokenize(code, annotations)

    @property
    def task_key(self) -> str:
        # TODO use which info as task key?
        # Task are supposed to be statefull, same input with same output
        return f"{type(self).__name__}-{self.task_hash}"

    def initialize_context(self):
        super().initialize_context()
        task_key = self.task_key
        self.config_dict.update({
            'task_key': task_key,
        })
        self.context.update({
            'task_key': task_key,
        })

    @property
    def task_workdir(self) -> Path:
        """Task's workdir is resolved in a bottom up way.
        In this hierarchical way, users can change flowrun.context['flow_workdir'] by setting up the flowrun's
        initial context, thus make the flow's workdir configurable.

        1: if the task_workdir is already absolute, then use it
        2: if parent_flow_workdir/task_workdir is already absolute, then use it
        3: otherwise use top_flow_workdir/parent_flow_workdir/task_workdir as the workdir

        Returns
        -------

        """
        workdir = Path(self.context['task_config']['workdir'], self.context['task_key'])
        if workdir.is_absolute():
            return workdir
        workdir = Path(self.context['flow_config']['workdir'], workdir)
        if workdir.is_absolute():
            return workdir
        workdir = Path(self.context['flow_workdir'], workdir)
        assert workdir.is_absolute()
        return workdir

    @property
    def run_workdir(self) -> Path:
        return Path(self.task_workdir, self.context['run_key'])

    def run_hash_source(self, run_data: BoundArguments, **kwargs) -> dict:
        return {
            'data': tuple(run_data.arguments.values())
        }

    @property
    def run_lock_source(self) -> List[str]:
        return [self.context.get('run_workdir')]

    async def call_run(self, data: BoundArguments, **kwargs) -> State:
        """Call self.run within the control of a asyncio.Lock identified by run_workdir
        Parameters
        ----------
        data
        kwargs

        Returns
        -------

        """
        from copy import copy
        # get input hash, we do not count for parameter names, we only care about orders
        # must use tuple, data.arguments.values() return an object instead of a container
        cache_type = self.context.get('cache_type', None)
        if cache_type:
            run_key: str = flowsaber.context.cache.hash(**self.run_hash_source(data, **kwargs))
        else:
            run_key: str = flowsaber.context.random_id
        assert run_key

        # create a fresh new task
        task = copy(self)
        task.context['run_key'] = run_key  # task.run_workdir need this
        context_update = {
            'run_key': run_key,
            'run_workdir': str(task.run_workdir)
        }
        # safe to update, flowsaber.context belongs to this run since we call handle_run_data
        task.context.update(context_update)
        flowsaber.context.update(context_update)
        # must lock _input key to avoid collision in cache and files in _running path
        async with flowsaber.context.lock(task.run_lock_source):
            task_runner = TaskRunner(
                task=task,
                inputs=data,
                server_address=self.context.get('server_address', None)
            )
            state = await flowsaber.context.executor.run(task_runner.run, **kwargs)
        return state

    async def handle_res(self, res):
        """Only push Success state result into _output channels. Some state may be skipped in case of
        Exceptions occurred within the task runner and thus return a Drop(Failure) state as a signal.
        Parameters
        ----------
        res
        """
        assert isinstance(res, Done), f"The result is {res}, should be a state of instance of Done"
        if isinstance(res, Success):
            await self.enqueue_res(res.result)
        elif isinstance(res, Failure):
            raise res.result
        # for Drop, just ignore it

    def need_skip(self, bound_args: BoundArguments) -> bool:
        """ Check if the _input can be directly passed into _output channels by predicate of user specified self.skip_fn
        Parameters
        ----------
        bound_args

        Returns
        -------

        """
        if self.skip_fn:
            return self.skip_fn(**bound_args.arguments)
        else:
            return False

    def skip(self, skip_fn: Callable):
        """A decorator/function exposed for users to specify skip function.

        Parameters
        ----------
        skip_fn
        """
        assert callable(skip_fn)
        self.skip_fn = skip_fn

    def clean(self):
        """Functions called after the execution of task. For example, Cache need to persist cached data.
        """
        pass


class Edge(object):
    """A edge represents a dependency between a channel and a task. the Task consumes data emited by the channel.
    """

    def __init__(self, channel: Channel, task: BaseTask):
        self.channel: Channel = channel
        self.task: BaseTask = task

    def serialize(self) -> EdgeInput:
        return EdgeInput(
            channel_id=self.channel.id,
            task_id=self.task.config_dict['id']
        )


run = class_deco(RunTask, 'run')
task = class_deco(Task, 'run')



"""
Expose all variables in core sub-package.
"""
# noinspection PyUnresolvedReferences
import flowsaber
# noinspection PyUnresolvedReferences
from flowsaber.core.channel import *
# noinspection PyUnresolvedReferences
from flowsaber.core.engine.flow_runner import *
# noinspection PyUnresolvedReferences
from flowsaber.core.engine.task_runner import *
# noinspection PyUnresolvedReferences
from flowsaber.core.flow import *
# noinspection PyUnresolvedReferences
from flowsaber.core.operators import *
# noinspection PyUnresolvedReferences
from flowsaber.core.task import *
# noinspection PyUnresolvedReferences
from flowsaber.core.task import *
# noinspection PyUnresolvedReferences
from flowsaber.core.utility.cache import *
# noinspection PyUnresolvedReferences
from flowsaber.core.utility.executor import *
# noinspection PyUnresolvedReferences
from flowsaber.core.utility.state import *
# noinspection PyUnresolvedReferences
from flowsaber.core.utility.target import *
# noinspection PyUnresolvedReferences
from flowsaber.core.utils import *
# noinspection PyUnresolvedReferences
from flowsaber.tasks import *


def run(flow: Flow, server_address: str = None,
        agent_id: str = None, context: dict = None,
        id: str = None, name: str = None, labels: list = None, **kwargs):
    """

    Parameters
    ----------
    flow
    server_address
    agent_id
    context
    id
    name
    labels
    kwargs

    Returns
    -------

    """
    from flowsaber.client.client import Client, GraphqlError

    assert flow.state == Flow.INITIALIZED, "The flow must be initialized and not being executed yet."
    context = context or {}
    with flowsaber.context(context) as context:
        merged_context = context.to_dict()

    if not agent_id:
        flowsaber.context.logger.info("Run the flow in local.")
        runner = FlowRunner(flow, server_address=server_address)
        runner.run(context=merged_context)
    else:
        assert server_address, "Must specify a server to schedule flowrun in remote agent."
        flowrun_input = FlowRunInput(
            id=id or flowsaber.context.random_id,
            name=name or flowsaber.context.random_id,
            labels=labels or [],
            context=merged_context,
            state=Scheduled().to_dict(),
            flow_id=flow.config_dict['id'],
            agent_id=agent_id,
            **kwargs
        )

        async def upload_can_run():
            client = Client(server_address)

            flowsaber.context.logger.info(f"Upload the flow onto the server: {server_address}")
            try:
                # test if the flow already exists in server.
                flow_data = await client.query("get_flow", flow.config.id, "id")
            except GraphqlError:
                await client.mutation("create_flow", flow.serialize(), "id")
                flow_data = await client.query("get_flow", flow.config.id, "id")

            flowsaber.context.logger.info(f"Scheduling the flow run in server: {server_address}")
            flowrun_data = await client.mutation("update_flowrun", flowrun_input, 'id name labels')
            flowsaber.context.logger.info(f"Scheduled the flow run: {flowrun_data} in server: {server_address} ")

            await client.close()
            return flowrun_data

        return asyncio.run(upload_can_run())



import os
import subprocess
from pathlib import Path
from typing import List, Union, Tuple, Optional, Any

from flowsaber.core.channel import Output
from flowsaber.core.flow import Flow
from flowsaber.core.task import Task, RunTask
from flowsaber.core.utility.target import Stdin, Stdout, File
from flowsaber.core.utils import class_deco
from flowsaber.utility.utils import change_cwd, capture_local


class ShellTaskExecuteError(RuntimeError):
    pass


class CommandTaskComposeError(RuntimeError):
    pass


class ShellTask(Task):
    """Task that execute bash command by using subprocess.
    """

    default_config = {
        'publish_dirs': [],
    }

    def run(self, cmd: str, output=None, envs: dict = None):
        """This method should be thread safe, can not run functions depends one process-base attributes like ENV, ....

        Parameters
        ----------
        cmd
        output
        envs: dict

        Returns
        -------

        """
        flow_workdir = self.context['flow_workdir']
        run_workdir = self.context['run_workdir']
        # run bash command in shell with Popen
        stdout_file, stderr_file = self.execute_shell_command(cmd, run_workdir, envs)
        # 1. return stdout simulated by file
        if output is None:
            stdout_path = Path(stdout_file).resolve()
            stdout = Stdout(stdout_path)
            stdout.initialize_hash()
            return stdout

        # 2. return globed files
        collect_files: List[File] = []
        resolved_output = self.glob_output_files(output, run_workdir, collect_files)
        publish_dirs = self.get_publish_dirs(flow_workdir, self.config_dict.get('publish_dirs', []))
        for file in collect_files:
            file.initialize_hash()
            for pub_dir in publish_dirs:
                pub_dir.mkdir(parents=True, exist_ok=True)
                pub_file = pub_dir / Path(file.name)
                if not pub_file.exists():
                    file.link_to(pub_file)
        return resolved_output

    def execute_shell_command(self, cmd: str, run_workdir: Union[str, Path], envs: dict = None) -> Tuple[Path, Path]:
        """Run command in shell

        Parameters
        ----------
        cmd
        run_workdir
        envs

        Returns
        -------

        """
        # this is not thread safe, because ENV is process-based attributes
        wrapped_cmd = f"mkdir -p {run_workdir};\n" \
                      f"cd {run_workdir};\n" \
                      f"{cmd}"
        # write into a sh file
        bash_script = Path(run_workdir, f".__run__.sh")
        with change_cwd(run_workdir):
            with bash_script.open('w') as f:
                f.write(wrapped_cmd)
        # execute with subprocess.Popen
        with change_cwd(run_workdir):
            envs = envs or os.environ.copy()
            stdout_f = run_workdir / Path(f".__run__.stdout")
            stderr_f = run_workdir / Path(f".__run__.stderr")
            with stdout_f.open('w') as stdout_file, stderr_f.open('w+') as stderr_file:
                with subprocess.Popen(f"bash -e {bash_script};",
                                      stdout=stdout_file, stderr=stderr_file,
                                      env=envs, shell=True) as p:
                    p.wait()

                stderr_file.seek(0)
                stderr = stderr_file.read(1000)
                if p.returncode:
                    raise ShellTaskExecuteError(f"Execute bash file: {bash_script} meets "
                                                f"error: {stderr} in {run_workdir}.")

            return stdout_f, stderr_f

    @classmethod
    def glob_output_files(cls, item, run_workdir, collect_files: List[File]):
        """Iterate over the item hierarchically and convert in-place and glob found str into Files
        and collect them into the third parameter.

        Parameters
        ----------
        item
        run_workdir
        collect_files

        Returns
        -------

        """
        if type(item) is str:
            # key step
            files = [File(p.resolve())
                     for p in Path(run_workdir).glob(item)
                     if not p.name.startswith('.') and p.is_file()]
            cls.glob_output_files(files, run_workdir, collect_files)
            # may globed multiple files
            return files[0] if len(files) == 1 else tuple(files)

        elif isinstance(item, dict):
            for k, v in item.items():
                item[k] = cls.glob_output_files(v, run_workdir, collect_files)
        elif isinstance(item, (tuple, list)):
            if isinstance(item, tuple):
                item = list(item)
            for i, v in enumerate(item):
                item[i] = cls.glob_output_files(v, run_workdir, collect_files)
        elif isinstance(item, File):
            # collect File
            collect_files.append(item)

        return item

    @staticmethod
    def get_publish_dirs(flow_workdir, configured_publish_dirs: List[str]):
        """Get absolute path of configured publish dirs.

        Parameters
        ----------
        flow_workdir
        configured_publish_dirs

        Returns
        -------

        """
        # resolve publish dirs
        publish_dirs = []
        for pub_dir in configured_publish_dirs:
            pub_dir = Path(pub_dir).expanduser().resolve()
            if pub_dir.is_absolute():
                publish_dirs.append(pub_dir)
            else:
                publish_dirs.append(Path(flow_workdir, pub_dir).resolve())
        return publish_dirs


class CommandTask(RunTask):
    """Task used for composing bash command based on outputs data of channels. Users need to implement the
    command method.
    Note that the _output Channel of this task simply emits composed bash command in str type, and this
    bash command needs to be actually executed by ShellTask.
    """
    FUNC_PAIRS = [('command', 'run')]

    def __init__(self, **kwargs):
        super().__init__(num_out=2, **kwargs)

    def command(self, *args, **kwargs) -> str:
        """Users need to implement this function to compose the final bash command.

        The returned value of this method represents the expected outputs after executing the
        composed bash command in shell:
            1. None represents the output is stdout.
            2. str variables represents glob syntax for files in the working directory.

        To tell flowsaber what's the composed bash command, users has two options:
            1: Assign the composed command to a variable named CMD.
            2: Write virtual fstring as the docstring of command method. All variables in the command method
                scoped can be used freely.

        Here are some examples:

            class A(CommandTask):
                def command(self, fa, fasta):
                    "bwa map -t {self.context.cpu} {fa} {fasta} -o {bam_file}"
                    bam_file = "test.bam"
                    return bam_file

            class B(CommandTask):
                def command(self, file):
                    a = "xxxx"
                    b = 'xxxx'
                    CMD = f"echo {a}\n"
                          f"echo {b}\n"
                          f"cat {file}"
                    # here implicitly returned a None, represents the _output of cmd is stdout

        Parameters
        ----------
        args
        kwargs
        """
        raise NotImplementedError("Please implement this function and return a bash script.")

    def compose_command(self, *args, **kwargs) -> Tuple[str, Any]:
        # TODO this cause error in debug mode, since settrace is forbidden in debug mode.
        # TWO options: 1: use local var CMD or use docstring
        with capture_local() as local_vars:
            cmd_output = self.command(*args, **kwargs)
        cmd = local_vars.get("CMD", None)
        if cmd is None:
            cmd = self.command.__doc__
            if cmd is None:
                raise CommandTaskComposeError("CommandTask must be registered with a shell commands by "
                                              "assigning the commands into a CMD variable inside the function or "
                                              "adding the commands as the command() method's documentation by "
                                              "setting `__doc__`.")
            local_vars.update({'self': self})
            cmd = cmd.format(**local_vars)
        # add cat stdin cmd
        stdins = [arg for arg in list(args) + list(kwargs.values()) if isinstance(arg, Stdin)]
        if len(stdins) > 1:
            raise CommandTaskComposeError(f"Found more than two stdin inputs: {stdins}")
        stdin = f"{stdins[0]} " if len(stdins) else ""
        cmd = f"{stdin}{cmd}"

        return cmd, cmd_output

    def run(self, *args, **kwargs):
        cmd, cmd_output = self.compose_command(*args, **kwargs)

        return cmd, cmd_output


class ShellFlow(Flow):
    FUNC_PAIRS = [('command', 'run')]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.command_task_: Optional[CommandTask] = None
        self.shell_task_: Optional[ShellTask] = None

    @property
    def command_task(self) -> CommandTask:
        if self.command_task_ is None:
            self.command_task_ = command(type(self).command, **self.rest_kwargs)
        return self.command_task_

    @property
    def shell_task(self) -> ShellTask:
        if self.shell_task_ is None:
            # rename shell_task's name to FlowNameShellTask
            flow_name = self.config_dict['name']
            if '[' in flow_name:
                flow_name = flow_name[:flow_name.index('[')]
            task_cls_name = flow_name + "ShellTask"
            task_cls_name = task_cls_name[0].upper() + task_cls_name[1:]
            self.shell_task_ = type(task_cls_name, (ShellTask,), {})(**self.rest_kwargs)
        return self.shell_task_

    def command(self, *args, **kwargs):
        raise NotImplementedError("Users need to implement this method to indicate flowsaber how to "
                                  "compose the bash command.")

    def run(self, *args, **kwargs) -> Output:
        cmd_ch, cmd_output_ch = self.command_task(*args, **kwargs)
        output_ch = self.shell_task(cmd_ch, cmd_output_ch)

        return output_ch


command = class_deco(CommandTask, 'command')
shell = class_deco(ShellFlow, 'command')
